# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Transport/operating_empty_weight.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ----------------------------------------------------------------------------------------------------------------------
import RCAIDE
from RCAIDE.Framework.Core import Data ,  Units 
from RCAIDE.Library.Methods.Weights.Correlation_Buildups import Propulsion as Propulsion 
from RCAIDE.Library.Methods.Weights.Correlation_Buildups import Common    as  Common
from RCAIDE.Library.Methods.Weights.Correlation_Buildups import Raymer    as  Raymer 
from RCAIDE.Library.Methods.Weights.Correlation_Buildups import FLOPS     as  FLOPS 
from RCAIDE.Library.Methods.Weights.Correlation_Buildups import Transport as  Transport  
from RCAIDE.Library.Attributes.Materials.Aluminum import Aluminum

# python imports 
import numpy as np

# ---------------------------------------------------------------------------------------------------------------------- 
# Operating Empty Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_operating_empty_weight(vehicle, settings=None, method_type='RCAIDE'):
    """ Main function that estimates the zero-fuel weight of a transport aircraft:
        - MTOW = WZFW + FUEL
        - WZFW = WOE + WPAYLOAD
        - WOE = WE + WOPERATING_ITEMS
        - WE = WSTRCT + WPROP + WSYS
        Assumptions:
            1) All nacelles are identical
            2) The number of nacelles is the same as the number of engines 

        Source:
            FLOPS method: The Flight Optimization System Weight Estimation Method
            RCAIDE method: http://aerodesign.stanford.edu/aircraftdesign/AircraftDesign.html
            RAYMER method: Aircraft Design A Conceptual Approach
       Inputs:
            vehicle - data dictionary with vehicle properties               [dimensionless]
                -.networks: data dictionary with all the network elements and properties
                    -.total_weight: total weight of the propulsion system   [kg] 
                      (optional, calculated if not included)
                -.fuselages: data dictionary with the fuselage properties of the vehicle
                -.wings: data dictionary with all the wing properties of the vehicle, including horzinotal and vertical stabilizers
                -.wings['main_wing']: data dictionary with main wing properties
                    -.flap_ratio: flap surface area over wing surface area
                -.mass_properties: data dictionary with all the main mass properties of the vehicle including MTOW, ZFW, EW and OEW

            settings.weight_reduction_factors.
                    main_wing                                               [dimensionless] (.1 is a 10% weight reduction)
                    empennage                                               [dimensionless] (.1 is a 10% weight reduction)
                    fuselage                                                [dimensionless] (.1 is a 10% weight reduction)
            method_type - weight estimation method chosen, available:
                            - FLOPS Simple
                            - FLOPS Complex
                            - RCAIDE 
                            - Raymer
       Outputs:
            output - data dictionary with the weight breakdown of the vehicle
                        -.structures: structural weight
                            -.wing: wing weight
                            -.horizontal_tail: horizontal tail weight
                            -.vertical_tail: vertical tail weight
                            -.fuselage: fuselage weight
                            -.main_landing_gear: main landing gear weight
                            -.nose_landing_gear: nose landing gear weight
                            -.nacelle: nacelle weight
                            -.paint: paint weight
                            -.total: total strucural weight

                        -.propulsion: propulsive system weight
                            -.engines: dry engine weight
                            -.thrust_reversers: thrust reversers weight
                            -.miscellaneous: miscellaneous items includes electrical system for engines and starter engine
                            -.fuel_system: fuel system weight
                            -.total: total propulsive system weight

                        -.systems: system weight
                            -.control_systems: control system weight
                            -.apu: apu weight
                            -.electrical: electrical system weight
                            -.avionics: avionics weight
                            -.hydraulics: hydraulics and pneumatic system weight
                            -.furnish: furnishing weight
                            -.air_conditioner: air conditioner weight
                            -.instruments: instrumentation weight
                            -.anti_ice: anti ice system weight
                            -.total: total system weight

                        -.payload: payload weight
                            -.passengers: passenger weight
                            -.bagage: baggage weight
                            -.cargo: cargo weight
                            -.total: total payload weight

                        -.operational_items: operational items weight
                            -.misc: unusable fuel, engine oil, passenger service weight and cargo containers
                            -.flight_crew: flight crew weight
                            -.flight_attendants: flight attendants weight
                            -.total: total operating items weight

                        -.empty = structures.total + propulsion.total + systems.total
                        -.operating_empty = empty + operational_items.total
                        -.zero_fuel_weight = operating_empty + payload.total
                        -.fuel = vehicle.mass_properties.max_takeoff - zero_fuel_weight


        Properties Used:
            N/A
    """
    
    if settings == None:
        W_factors = Data()
        use_max_fuel_weight = True 
    else:
        use_max_fuel_weight = settings.use_max_fuel_weight 
        
    # Set the factors
    if not hasattr(settings, 'weight_reduction_factors'):
        W_factors              = Data() 
        W_factors.main_wing    = 0.
        W_factors.empennage    = 0.
        W_factors.fuselage     = 0.
        W_factors.structural   = 0.
        W_factors.systems      = 0.
    else:
        W_factors = settings.weight_reduction_factors
        if 'structural' in W_factors and W_factors.structural != 0.:
            print('Overriding individual structural weight factors')
            W_factors.main_wing    = 0.
            W_factors.empennage    = 0.
            W_factors.fuselage     = 0.
            W_factors.systems      = 0.
        else:
            W_factors.structural   = 0.
            W_factors.systems      = 0. 
    
    Wings = RCAIDE.Library.Components.Wings  
    if method_type == 'FLOPS Simple' or method_type == 'FLOPS Complex':
        if vehicle.flight_envelope.design_mach_number  == None: # Added design mach number
            raise ValueError("FLOPS requires a design mach number for sizing!")
        if vehicle.flight_envelope.design_range  == None:
            raise ValueError("FLOPS requires a design range for sizing!")
        if vehicle.flight_envelope.design_cruise_altitude == None:
            raise ValueError("FLOPS requires a cruise altitude for sizing!")
        if not hasattr(vehicle, 'flap_ratio'):
            if vehicle.systems.accessories == "sst":
                flap_ratio = 0.22
            else:
                flap_ratio = 0.33
            for wing in vehicle.wings:
                if isinstance(wing, Wings.Main_Wing):
                    wing.flap_ratio = flap_ratio 
                
    ##-------------------------------------------------------------------------------             
    # Payload Weight
    ##-------------------------------------------------------------------------------  
    if method_type == 'FLOPS Simple' or method_type == 'FLOPS Complex':
        payload = FLOPS.compute_payload_weight(vehicle)
    else:
        payload = Common.compute_payload_weight(vehicle)
    
    
    vehicle.payload.passengers                      = RCAIDE.Library.Components.Component()
    vehicle.payload.passengers.tag                  = 'passengers'
    vehicle.payload.passengers.mass_properties.mass = payload.passengers
    
    vehicle.payload.baggage                         = RCAIDE.Library.Components.Component()
    vehicle.payload.baggage.tag                     = 'baggage'
    vehicle.payload.baggage.mass_properties.mass    = payload.baggage
    
    vehicle.payload.cargo                           = RCAIDE.Library.Components.Component() 
    vehicle.payload.cargo.tag                       = 'cargo'   
    vehicle.payload.cargo.mass_properties.mass      = payload.cargo    

    ##-------------------------------------------------------------------------------             
    # Operating Items Weight
    ##------------------------------------------------------------------------------- 
    if method_type == 'FLOPS Simple' or method_type == 'FLOPS Complex':
        W_oper = FLOPS.compute_operating_items_weight(vehicle)
    else:
        W_oper = Transport.compute_operating_items_weight(vehicle)  

    ##-------------------------------------------------------------------------------         
    # System Weight
    ##------------------------------------------------------------------------------- 
    if method_type == 'FLOPS Simple' or method_type == 'FLOPS Complex':
        W_systems = FLOPS.compute_systems_weight(vehicle)
    elif method_type == 'Raymer':
        W_systems = Raymer.compute_systems_weight(vehicle)
    else:
        W_systems = Common.compute_systems_weight(vehicle)
        
    for item in W_systems.keys():
        W_systems[item] *= (1. - W_factors.systems)
    
    ##-------------------------------------------------------------------------------                 
    # Propulsion Weight 
    ##-------------------------------------------------------------------------------
    output                                      = Data()
    output.empty                                = Data() 
    output.empty.propulsion                     = Data() 
    output.empty.propulsion.total               = 0
    output.empty.propulsion.engines             = 0
    output.empty.propulsion.thrust_reversers    = 0
    output.empty.propulsion.miscellaneous       = 0
    output.empty.propulsion.fuel_system         = 0

    W_energy_network                   = Data()
    W_energy_network.total             = 0
    W_energy_network.W_engine          = 0 
    W_energy_network.W_thrust_reverser = 0 
    W_energy_network.W_engine_controls = 0 
    W_energy_network.W_starter         = 0 
    W_energy_network.W_fuel_system     = 0 
    W_energy_network.W_motors          = 0 
    W_energy_network.W_nacelle         = 0 
    W_energy_network.W_battery         = 0
    W_energy_network.W_motor           = 0
    number_of_engines                  = 0
    number_of_tanks                    = 0
    W_energy_network_cumulative        = 0 

    for network in vehicle.networks: 
        W_energy_network_total   = 0 
        # Fuel-Powered Propulsors  
        if method_type == 'FLOPS Simple' or method_type == 'FLOPS Complex':
            W_propulsion                         = FLOPS.compute_propulsion_system_weight(vehicle, network)
            W_energy_network_total              += W_propulsion.W_prop 
            W_energy_network.W_engine           += W_propulsion.W_engine
            W_energy_network.W_thrust_reverser  += W_propulsion.W_thrust_reverser
            W_energy_network.W_engine_controls  += W_propulsion.W_engine_controls
            W_energy_network.W_starter          += W_propulsion.W_starter
            W_energy_network.W_fuel_system      += W_propulsion.W_fuel_system 
            W_energy_network.W_nacelle          += W_propulsion.W_nacelle    
            number_of_engines                   += W_propulsion.number_of_engines
            number_of_tanks                     += W_propulsion.number_of_fuel_tanks  
            for propulsor in network.propulsors:
                propulsor.mass_properties.mass = W_energy_network_total / number_of_engines
            
        elif method_type == 'Raymer':
            W_propulsion                        = Raymer.compute_propulsion_system_weight(vehicle, network) 
            W_energy_network_total              += W_propulsion.W_prop 
            W_energy_network.W_engine           += W_propulsion.W_engine
            W_energy_network.W_thrust_reverser  += W_propulsion.W_thrust_reverser
            W_energy_network.W_engine_controls  += W_propulsion.W_engine_controls
            W_energy_network.W_starter          += W_propulsion.W_starter
            W_energy_network.W_fuel_system      += W_propulsion.W_fuel_system 
            W_energy_network.W_nacelle          += W_propulsion.W_nacelle    
            number_of_engines                   += W_propulsion.number_of_engines
            number_of_tanks                     += W_propulsion.number_of_fuel_tanks
            for propulsor in network.propulsors:
                propulsor.mass_properties.mass = W_energy_network_total / number_of_engines
        else:
            number_of_tanks = 0
            for fuel_line in  network.fuel_lines: 
                for fuel_tank in fuel_line.fuel_tanks:
                    number_of_tanks +=  1
            for propulsor in network.propulsors:
                if isinstance(propulsor, RCAIDE.Library.Components.Propulsors.Turbofan):
                    thrust_sls                     = propulsor.sealevel_static_thrust  
                    W_engine_jet                   = Propulsion.compute_jet_engine_weight(thrust_sls)
                    total_propulsor_mass           = Propulsion.integrated_propulsion(W_engine_jet) 
                    propulsor.mass_properties.mass = total_propulsor_mass 
                    W_energy_network_total         += total_propulsor_mass 
                    number_of_engines += 1             
                 
        # Electric-Powered Propulsors  
        for bus in network.busses: 
            # electrical payload 
            W_systems.W_electrical  += bus.payload.mass_properties.mass * Units.kg
     
            # Avionics Weight 
            W_systems.W_avionics  += bus.avionics.mass_properties.mass      
    
            for battery in bus.battery_modules: 
                W_energy_network_total  += battery.mass_properties.mass * Units.kg
                W_energy_network.W_battery = battery.mass_properties.mass * Units.kg
                
        for propulsor in network.propulsors:
            if 'motor' in propulsor:                           
                W_energy_network.W_motor +=  propulsor.motor.mass_properties.mass
                W_energy_network_total  +=  propulsor.motor.mass_properties.mass
                   
    W_energy_network_cumulative += W_energy_network_total
    
    ##-------------------------------------------------------------------------------                 
    # Pod Weight Weight 
    ##-------------------------------------------------------------------------------         
    WPOD  = 0.0             
    if method_type == 'FLOPS Complex': 
        NENG   = number_of_engines
        WTNFA  = W_energy_network.W_engine + W_energy_network.W_thrust_reverser + W_energy_network.W_starter \
                + 0.25 * W_energy_network.W_engine_controls + 0.11 * W_systems.W_instruments + 0.13 * W_systems.W_electrical \
                + 0.13 * W_systems.W_hyd_pnu + 0.25 * W_energy_network.W_fuel_system
        WPOD += WTNFA / np.max([1, NENG]) + W_energy_network.W_nacelle / np.max(
            [1.0, NENG + 1. / 2 * (NENG - 2 * np.floor(NENG / 2.))])
 
    output.empty.propulsion.total               = W_energy_network_cumulative
    output.empty.propulsion.battery             = W_energy_network.W_battery
    output.empty.propulsion.motors              = W_energy_network.W_motor
    output.empty.propulsion.engines             = W_energy_network.W_engine
    output.empty.propulsion.thrust_reversers    = W_energy_network.W_thrust_reverser
    output.empty.propulsion.miscellaneous       = W_energy_network.W_engine_controls + W_energy_network.W_starter
    output.empty.propulsion.fuel_system         = W_energy_network.W_fuel_system

    ##-------------------------------------------------------------------------------                 
    # Wing Weight 
    ##------------------------------------------------------------------------------- 
    Al_rho   = Aluminum().density
    Al_sigma = Aluminum().yield_tensile_strength      
    
    num_main_wings      = 0
    W_main_wing        = 0.0
    W_tail_horizontal  = 0.0
    W_tail_vertical    = 0.0
    for wing in vehicle.wings:
        if isinstance(wing, Wings.Main_Wing):
            num_main_wings += 1
    
    for wing in vehicle.wings:
        if isinstance(wing, Wings.Main_Wing): 
            if method_type == 'FLOPS Simple' or method_type == 'FLOPS Complex':
                complexity = method_type.split()[1]
                W_wing = FLOPS.compute_wing_weight(vehicle, wing, WPOD, complexity, settings, num_main_wings)
            elif method_type == 'Raymer':
                W_wing = Raymer.compute_main_wing_weight(vehicle, wing) 
            else:
                W_wing = Common.compute_main_wing_weight(vehicle, wing, Al_rho, Al_sigma) 
            # Apply weight factor
            W_wing = W_wing * (1. - W_factors.main_wing) * (1. - W_factors.structural)
            if np.isnan(W_wing):
                W_wing = 0.
            wing.mass_properties.mass = W_wing
            W_main_wing += W_wing
        if isinstance(wing, Wings.Horizontal_Tail):
            if method_type == 'FLOPS Simple' or method_type == 'FLOPS Complex':
                W_tail = FLOPS.compute_horizontal_tail_weight(vehicle, wing)
            elif method_type == 'Raymer':
                W_tail = Raymer.compute_horizontal_tail_weight(vehicle, wing)
            else:
                W_tail = Transport.compute_horizontal_tail_weight(vehicle, wing)
            if type(W_tail) == np.ndarray:
                W_tail = sum(W_tail)
            # Apply weight factor
            W_tail = W_tail * (1. - W_factors.empennage) * (1. - W_factors.structural)
            # Pack and sum
            wing.mass_properties.mass = W_tail
            W_tail_horizontal += W_tail
        if isinstance(wing, Wings.Vertical_Tail):
            if method_type == 'FLOPS Simple' or method_type == 'FLOPS Complex':
                W_tail = FLOPS.compute_vertical_tail_weight(vehicle, wing)
            elif method_type == 'Raymer':
                W_tail = Raymer.compute_vertical_tail_weight(vehicle, wing)
            else:
                W_tail = Transport.compute_vertical_tail_weight(vehicle, wing)
            # Apply weight factor
            W_tail = W_tail * (1. - W_factors.empennage) * (1. - W_factors.structural)
            # Pack and sum
            wing.mass_properties.mass = W_tail
            W_tail_vertical += W_tail
        
    ##-------------------------------------------------------------------------------                 
    # Fuselage 
    ##------------------------------------------------------------------------------- 
    W_fuselage_total = 0
    for fuse in vehicle.fuselages:
        if method_type == 'FLOPS Simple' or method_type == 'FLOPS Complex':
            W_fuselage = FLOPS.compute_fuselage_weight(vehicle)
        elif method_type == 'Raymer':
            W_fuselage = Raymer.compute_fuselage_weight(vehicle, fuse, settings)
        else:
            W_fuselage = Transport.compute_fuselage_weight(vehicle, fuse, W_main_wing, W_energy_network_cumulative)
        W_fuselage = W_fuselage * (1. - W_factors.fuselage) * (1. - W_factors.structural)
        fuse.mass_properties.mass = W_fuselage
        W_fuselage_total += W_fuselage
    
    ##-------------------------------------------------------------------------------                 
    # Landing Gear Weight
    ##------------------------------------------------------------------------------- 
    if method_type == 'FLOPS Simple' or method_type == 'FLOPS Complex':
        landing_gear = FLOPS.compute_landing_gear_weight(vehicle)
    elif method_type == 'Raymer':
        landing_gear = Raymer.compute_landing_gear_weight(vehicle)
    else:
        landing_gear =  Common.compute_landing_gear_weight(vehicle) 
    
    ##-------------------------------------------------------------------------------                 
    # Accumulate Structural Weight
    ##-------------------------------------------------------------------------------   
    output.empty.structural                       = Data()
    output.empty.structural.wings                  = W_main_wing +   W_tail_horizontal +  W_tail_vertical 
    output.empty.structural.fuselage              = W_fuselage_total
    output.empty.structural.landing_gear          = landing_gear.main +  landing_gear.nose  
    output.empty.structural.nacelle               = W_energy_network.W_nacelle
    
    if 'FLOPS' in method_type:
        print('Paint weight is currently ignored in FLOPS calculations.')
    output.empty.structural.paint = 0  # TODO reconcile FLOPS paint calculations with Raymer and RCAIDE baseline
    output.empty.structural.total = output.empty.structural.wings   + output.empty.structural.fuselage + output.empty.structural.landing_gear\
                              + output.empty.structural.paint + output.empty.structural.nacelle 

    ##-------------------------------------------------------------------------------                 
    # Accumulate Systems Weight
    ##-------------------------------------------------------------------------------
    output.empty.systems                        = Data()
    output.empty.systems.control_systems        = W_systems.W_flight_control
    output.empty.systems.apu                    = W_systems.W_apu
    output.empty.systems.electrical             = W_systems.W_electrical
    output.empty.systems.avionics               = W_systems.W_avionics
    output.empty.systems.hydraulics             = W_systems.W_hyd_pnu
    output.empty.systems.furnishings            = W_systems.W_furnish
    output.empty.systems.air_conditioner        = W_systems.W_ac + W_systems.W_anti_ice # Anti-ice is sometimes included in ECS
    output.empty.systems.instruments            = W_systems.W_instruments
    output.empty.systems.total                  = output.empty.systems.control_systems + output.empty.systems.apu \
                                                    + output.empty.systems.electrical + output.empty.systems.avionics \
                                                    + output.empty.systems.hydraulics + output.empty.systems.furnishings \
                                                    + output.empty.systems.air_conditioner + output.empty.systems.instruments
 
    output.payload    = payload 
    output.operational_items    = Data()
    output.operational_items    = W_oper 
    output.empty.total          = output.empty.structural.total + output.empty.propulsion.total + output.empty.systems.total 
    output.zero_fuel_weight     = output.empty.total + output.operational_items.total + output.payload.total
    output.max_takeoff          = vehicle.mass_properties.max_takeoff
    total_fuel_weight           = vehicle.mass_properties.max_takeoff - output.zero_fuel_weight
    
    # assume fuel is equally distributed in fuel tanks
    if use_max_fuel_weight:
        for network in vehicle.networks: 
            for fuel_line in network.fuel_lines:  
                for fuel_tank in fuel_line.fuel_tanks:
                    fuel_weight =  total_fuel_weight/number_of_tanks  
                    fuel_tank.fuel.mass_properties.mass = fuel_weight
                    
    nose_landing_gear = False
    main_landing_gear =  False
    for LG in vehicle.landing_gears:
        if isinstance(LG, RCAIDE.Library.Components.Landing_Gear.Main_Landing_Gear):
            LG.mass_properties.mass = landing_gear.main
            main_landing_gear = True
        elif isinstance(LG, RCAIDE.Library.Components.Landing_Gear.Nose_Landing_Gear):
            LG.mass_properties.mass = landing_gear.nose
            nose_landing_gear = True 
    if nose_landing_gear == False:
        nose_gear = RCAIDE.Library.Components.Landing_Gear.Nose_Landing_Gear()  
        nose_gear.mass_properties.mass = landing_gear.nose    
        vehicle.append_component(nose_gear) 
    if main_landing_gear == False:
        main_gear = RCAIDE.Library.Components.Landing_Gear.Main_Landing_Gear()  
        main_gear.mass_properties.mass = landing_gear.main  
        vehicle.append_component(main_gear) 

    control_systems                         = RCAIDE.Library.Components.Component()
    control_systems.tag                     = 'control_systems'  
    control_systems.mass_properties.mass    = output.empty.systems.control_systems
    electrical_systems                      = RCAIDE.Library.Components.Component()
    electrical_systems.tag                  = 'electrical_systems'
    electrical_systems.mass_properties.mass = output.empty.systems.electrical
    furnishings                             = RCAIDE.Library.Components.Component()
    furnishings.tag                         = 'furnishings'
    furnishings.mass_properties.mass        = output.empty.systems.furnishings
    air_conditioner                         = RCAIDE.Library.Components.Component() 
    air_conditioner.tag                     = 'air_conditioner'
    air_conditioner.mass_properties.mass    = output.empty.systems.air_conditioner
    apu                                     = RCAIDE.Library.Components.Component()
    apu.tag                                 = 'apu'
    apu.mass_properties.mass                = output.empty.systems.apu
    hydraulics                              = RCAIDE.Library.Components.Component()
    hydraulics.tag                          = 'hydraulics' 
    hydraulics.mass_properties.mass         = output.empty.systems.hydraulics
    avionics                                = RCAIDE.Library.Components.Systems.Avionics()
    avionics.mass_properties.mass           = output.empty.systems.avionics + output.empty.systems.instruments
    optionals                               = RCAIDE.Library.Components.Component()
    optionals.tag                           = 'optionals'
    optionals                               = RCAIDE.Library.Components.Component()
    optionals.mass_properties.mass          = output.operational_items.misc
    
    # assign components to vehicle
    vehicle.systems.control_systems         = control_systems
    vehicle.systems.electrical_systems      = electrical_systems
    vehicle.systems.avionics                = avionics
    vehicle.systems.furnishings             = furnishings
    vehicle.systems.air_conditioner         = air_conditioner 
    vehicle.systems.apu                     = apu
    vehicle.systems.hydraulics              = hydraulics
    vehicle.systems.optionals               = optionals   

    return output



# RCAIDE/Library/Plots/Performance/Aerodynamics/plot_airfoil_boundary_layer_properties.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style

# python imports 
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------     

def plot_airfoil_boundary_layer_properties(ap,
                                           save_figure = False,
                                           show_legend = False,
                                           file_type = ".png",
                                           save_filename = 'Airfoil_with_Boundary_Layers', 
                                           width = 11, height = 7):
    """Plots viscous distributions
    
    Assumptions:
    None
    
    Source: 
    None
                                                     
    Inputs:
        ap     : data stucture of airfoil boundary layer properties  
                                                                           
    Outputs:
        Figures of quantity distributions
    
    Properties Used:
    N/A
    """      
    # get plotting style 
    ps      = plot_style()  
    
    plot_quantity(ap, ap.Ue_Vinf, r'U_e/U_inf$'  ,'Inviscid Edge Velocity',0,3,file_type,show_legend,save_figure,width,height) 
    plot_quantity(ap, ap.H,  r'H'  ,'Kinematic Shape Parameter',-1,10,file_type,show_legend,save_figure,width,height) 
    plot_quantity(ap, ap.delta_star, r'delta*' ,'Displacement Thickness',-0.01,0.1 ,file_type,show_legend,save_figure,width,height) 
    plot_quantity(ap, ap.delta   , r'delta' ,'Boundary Layer Thickness',-0.01,0.1 ,file_type,show_legend,save_figure,width,height) 
    plot_quantity(ap, ap.theta, r'theta' ,'Momentum Thickness',-0.001, 0.015,file_type,show_legend,save_figure,width,height) 
    plot_quantity(ap, ap.cf, r'c_f'  ,   'Skin Friction Coefficient',-0.1,1,file_type,show_legend,save_figure,width,height) 
    plot_quantity(ap, ap.Re_theta,  r'Re_theta'  ,'Theta Reynolds Number',-2E2,1E3,file_type,show_legend,save_figure,width,height)  
 
    n_cpts   = len(ap.AoA[:,0])
    n_cases  = len(ap.AoA[0,:]) 

    # create array of colors for difference reynolds numbers        
    blues = cm.winter(np.linspace(0,0.9,n_cases))     
    reds  = cm.autumn(np.linspace(0,0.9,n_cases))   

    fig   = plt.figure(save_filename)
    fig.set_size_inches(width,height)
    
    for i in range(n_cpts):   
        for j in range(n_cases):  
            axis_0 = plt.subplot(1,1,1) 
            axis_0.plot(ap.x[i,j], ap.y[i,j], color = blues[j], marker = ps.markers[0], linewidth = ps.line_width )
            axis_0.plot(ap.x[i,j][:-1], ap.y_bl[i,j], color = reds[j], marker = ps.markers[0], linewidth = ps.line_width ) 
            set_axes(axis_0)    
   
    # set title of plot 
    title_text    = 'Airfoil with Boundary Layers'  
    fig.suptitle(title_text)
    
    if save_figure:
        plt.savefig(save_filename + file_type)   
 
    return fig    
 
# ----------------------------------------------------------------------
#  Plot Quantity
# ----------------------------------------------------------------------  
def plot_quantity(ap, q, qaxis, qname,ylim_low,ylim_high,file_type,show_legend,save_figure,width,height) :
    """Plots a quantity q over lower/upper/wake surfaces
    
    Assumptions:
    None
    
    Source: 
    None
                                                     
    Inputs:
       ap        : data stucture of airfoil boundary layer properties  
       q         : vector of values to plot, on all points (wake too if present)
       qaxis     : name of quantity, for axis labeling
       qname     : name of quantity, for title labeling
                                                                           
    Outputs:
       Figure showing q versus x
    
    Properties Used:
    N/A
    """          

    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
    
    n_cpts   = len(ap.AoA[:,0])
    n_cases  = len(ap.AoA[0,:]) 
    
    fig   = plt.figure(qname.replace(" ", "_"))
    fig.set_size_inches(width,height) 
    axis  = fig.add_subplot(1,1,1)   
    
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,n_cases))      
    
    for i in range(n_cpts):   
        for j in range(n_cases): 
            case_label = 'AoA: ' + str(round(ap.AoA[i,j]/Units.degrees, 2)) + ', Re: ' + str(ap.Re[i,j]) 
            axis.plot( ap.x[i,j], q[i,j], color = line_colors[j], marker = ps.markers[0], linewidth = ps.line_width,  label =case_label)  
            axis.set_ylim([ylim_low,ylim_high]) 
     
    if show_legend:
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4) 
        
        # Adjusting the sub-plots for legend 
        fig.subplots_adjust(top=0.8)
        
    # set title of plot 
    title_text    = qname    
    fig.suptitle(title_text)
            
    if save_figure:
        plt.savefig(qname.replace(" ", "_") + file_type) 
          
    return  
  
   
           


# RCAIDE/Library/Methods/Thermal_Management/Batteries/Air_Cooled/append_air_cooling_conditions.py
# 
# Created:  Aug 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
import RCAIDE
from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_wavy_channel_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_wavy_channel_conditions(wavy_channel,segment,coolant_line):
     """ Packs the initial wavy channel heat acquisition conditions

         Assumptions:
         None

         Source:
         N/A

         Inputs:  
              wavy_channel          (data structure)              [None]

         Outputs:
             segment
               heat_removed                                       [watts]           
               outlet_coolant_temperature                         [Kelvin]
               coolant_mass_flow_rate                             [kg/s]
               effectiveness                                      [None]
               power                                              [watts]    
               
         Properties Used:
         None
     """             
     
     atmosphere    = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
     alt           = -segment.conditions.frames.inertial.position_vector[:,2] 
     if segment.temperature_deviation != None:
          temp_dev = segment.temperature_deviation    
     atmo_data    = atmosphere.compute_values(altitude = alt,temperature_deviation=temp_dev)
     
     
     ones_row                                                                                        = segment.state.ones_row
     segment.state.conditions.energy[coolant_line.tag][wavy_channel.tag]                            = Conditions()
     segment.state.conditions.energy[coolant_line.tag][wavy_channel.tag].heat_removed               = 0 * ones_row(1) 
     segment.state.conditions.energy[coolant_line.tag][wavy_channel.tag].outlet_coolant_temperature = atmo_data.temperature[0,0]  * ones_row(1)     
     segment.state.conditions.energy[coolant_line.tag][wavy_channel.tag].coolant_mass_flow_rate     = 0 * ones_row(1)  
     segment.state.conditions.energy[coolant_line.tag][wavy_channel.tag].effectiveness              = 0 * ones_row(1)
     segment.state.conditions.energy[coolant_line.tag][wavy_channel.tag].power                      = 0 * ones_row(1)
     
     return

def append_wavy_channel_segment_conditions(wavy_channel,segment,coolant_line,conditions):
     """Sets the initial wavy channel heat acquisition properties at the
        start of each segment as the last point from the previous segment 

         Assumptions:
         None

         Source:
         N/A

         Inputs:  
          wavy_channel          (data structure)              [None]

         Outputs:
         None

         Properties Used:
         None
     """     

     wavy_channel_conditions = conditions[coolant_line.tag][wavy_channel.tag]
     if segment.state.initials:  
          wavy_channel_initials                                    = segment.state.initials.conditions.energy[coolant_line.tag][wavy_channel.tag]
          wavy_channel_conditions.outlet_coolant_temperature[:,0] = wavy_channel_initials.outlet_coolant_temperature[-1,0]
          wavy_channel_conditions.power[:,0]                      = wavy_channel_initials.power[-1,0] 
          wavy_channel_conditions.coolant_mass_flow_rate[:,0]     = wavy_channel_initials.coolant_mass_flow_rate[-1,0]     
     
     return


# RCAIDE/Library/Methods/Propulsors/Converters/Combustor/compute_rayleigh.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------     

# package imports
import numpy as np 
from   RCAIDE.Library.Methods.Gas_Dynamics  import rayleigh, fm_solver

# ---------------------------------------------------------------------------------------------------------------------- 
# compute_rayleigh
# ----------------------------------------------------------------------------------------------------------------------    
def compute_rayleigh(combustor,combustor_conditions,conditions):
    """ This combutes the temperature and pressure change across the combustor using Rayleigh Line flow.
    The following properties are computed 
        combustor_conditions.outputs.
          stagnation_temperature             (numpy.ndarray): [K]  
          stagnation_pressure                (numpy.ndarray): [Pa]
          stagnation_enthalpy                (numpy.ndarray): [J/kg]
          fuel_to_air_ratio                  (numpy.ndarray): [unitless] 
   
    Assumptions:
        1. Constant efficiency and pressure ratio
        2. Limits temepature ratio if thermal choking occures 

    Source:
        https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/

    Args:
        conditions.freestream.
          isentropic_expansion_factor        (numpy.ndarray): [unitless]
          specific_heat_at_constant_pressure (numpy.ndarray): [J/(kg K)]
          temperature                        (numpy.ndarray): [K]
          stagnation_temperature             (numpy.ndarray): [K]
        combustor_conditions.inputs. 
          stagnation_temperature             (numpy.ndarray): [K]
          stagnation_pressure                (numpy.ndarray): [Pa]
        combustor.
          turbine_inlet_temperature                  (float): [K]
          pressure_ratio                             (float): [unitless]
          efficiency                                 (float): [unitless]
          area_ratio                                 (float): [unitless]
          fuel_data.specific_energy                  (float): [J/kg]
 
    Returns:
        None 
    """          
    # unpacking the values from conditions
    gamma  = conditions.freestream.isentropic_expansion_factor 
    Cp     = conditions.freestream.specific_heat_at_constant_pressure 
    
    # unpack properties of combustor 
    Tt_in  = combustor_conditions.inputs.stagnation_temperature
    Pt_in  = combustor_conditions.inputs.stagnation_pressure
    Mach   = combustor_conditions.inputs.mach_number
    Tt4    = combustor.turbine_inlet_temperature 
    eta_b  = combustor.efficiency 
    htf    = combustor.fuel_data.specific_energy
    ar     = combustor.area_ratio 

    # Isentropic decceleration through divergent nozzle
    Mach   = np.atleast_2d(fm_solver(ar,Mach[:,0],gamma[:,0])).T
    
    # Determine max stagnation temperature to thermally choke flow                                     
    Tt4_ray = Tt_in*(1.+gamma*Mach*Mach)**2./((2.*(1.+gamma)*Mach*Mach)*(1.+(gamma-1.)/2.*Mach*Mach))

    # Rayleigh limitations define Tt4, taking max temperature before choking
    Tt4                 = np.ones_like(Tt4_ray) * Tt4 
    Tt4[Tt4_ray <= Tt4] = Tt4_ray[Tt4_ray <= Tt4]
    
    # Rayleigh calculations
    M_out                = np.zeros_like(Pt_in)
    Ptr                  = np.zeros_like(Pt_in)
    M_out[:,0], Ptr[:,0] = rayleigh(gamma[:,0],Mach[:,0],Tt4[:,0]/Tt_in[:,0]) 
    Pt_out               = Ptr*Pt_in
         
    # Compute the stagnation enthalpies from stagnation temperatures
    ht4     = Tt4 * Cp 
    ht_in   = Tt_in * Cp 
    
    # Using the Turbine exit temperature, the fuel properties and freestream temperature 
    # to compute the fuel to air ratio f
    f       = (ht4 - ht_in)/(eta_b*htf-ht4)

    # Computing the exit static and stagnation conditions
    ht_out  = Tt4 * Cp 
    
    # Pack results 
    combustor_conditions.outputs.stagnation_temperature  = Tt4
    combustor_conditions.outputs.stagnation_pressure     = Pt_out
    combustor_conditions.outputs.stagnation_enthalpy     = ht_out
    combustor_conditions.outputs.fuel_to_air_ratio       = f    
    combustor_conditions.outputs.mach_number             = M_out
    return 



# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Raymer/compute_systems_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import RCAIDE 
from RCAIDE.Framework.Core    import Units, Data 

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
# Systems Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_systems_weight(vehicle):
    """ Calculates the system weight based on the Raymer method

        Assumptions:
            Number of flight control systems = 4
            Max APU weight = 70 lbs
            Assuming not a reciprocating engine and not a turboprop
            System Electrical Rating: 60 kv · A (typically 40-60 for transports, 110-160 for fighters & bombers)
            Uninstalled Avionics weight: 1400 lb (typically= 800-1400 lb)

        Source:
            Aircraft Design: A Conceptual Approach (2nd edition)

        Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.networks: data dictionary containing all propulsion properties
                -.number_of_engines: number of engines
                -.sealevel_static_thrust: thrust at sea level               [N]
                -.fuselages['fuselage'].lengths.total: fuselage total length    [meters]
                -.fuselages['fuselage'].width: fuselage width                   [meters]
                -.fuselages['fuselage'].heights.maximum: fuselage maximum height[meters]
                -.mass_properties.max_takeoff: MTOW                             [kilograms]
                -.design_mach_number: design mach number for cruise flight
                -.design_range: design range of aircraft                        [nmi]
                -.passengers: number of passengers in aircraft
                -.wings['main_wing']: data dictionary with main wing properties
                    -.sweeps.quarter_chord: quarter chord sweep                 [deg]
                    -.areas.reference: wing surface area                        [m^2]
                    -.spans.projected: projected span of wing                   [m]
                    -.flap_ratio: flap surface area over wing surface area
                -.payload: payload weight of aircraft                           [kg]

        Outputs:
            output - a data dictionary with fields:
               W_flight_controls - weight of the flight control system                                [kilograms]
               W_apu - weight of the apu                                                       [kilograms]
               W_hyd_pnu - weight of the hydraulics and pneumatics                             [kilograms]
               W_instruments - weight of the instruments and navigational equipment            [kilograms]
               W_avionics - weight of the avionics                                             [kilograms]
               W_electrical - weight of the electrical items                                         [kilograms]
               W_ac - weight of the air conditioning and anti-ice system                       [kilograms]
               W_furnish - weight of the furnishings in the fuselage                           [kilograms]
               W_anti_ice - weight of anti-ice system                                          [kilograms]

        Properties Used:
            N/A
    """
    flap_area = 0
    ref_wing  = None 
    for wing in  vehicle.wings:
        if isinstance(wing, RCAIDE.Library.Components.Wings.Main_Wing):
            ref_wing  = wing
            rc        = wing.chords.root
            taper     = wing.taper
            for cs in wing.control_surfaces:
                if type(cs) == RCAIDE.Library.Components.Wings.Control_Surfaces.Flap: 
                    sfs  = cs.span_fraction_start    
                    sfe  = cs.span_fraction_end       
                    c_f  = cs.chord_fraction
                    span = (sfe - sfs) *wing.spans.projected 
                    y1s = c_f*(rc -  taper *sfs * rc)
                    y2e = c_f*(rc -  taper *sfe * rc)
                    flap_area =  span * ( y1s + y2e) /2 
    S = 0
    if ref_wing == None:
        for wing in  vehicle.wings:
            if S < wing.areas.reference:
                ref_wing = wing
                rc        = wing.chords.root
                taper     = wing.taper
                for cs in wing.control_surfaces:
                    if type(cs) == RCAIDE.Library.Components.Wings.Control_Surfaces.Flap: 
                        sfs  = cs.span_fraction_start    
                        sfe  = cs.span_fraction_end       
                        c_f  = cs.chord_fraction
                        span =  (sfe - sfs) *wing.spans.projected 
                        y1s = c_f*(rc -  taper *sfs * rc)
                        y2e = c_f*(rc -  taper *sfe * rc)
                        flap_area =  span * ( y1s + y2e) /2 
    L_fus = 0
    for fuselage in vehicle.fuselages:
        if L_fus < fuselage.lengths.total:
            ref_fuselage = fuselage
    
    flap_ratio     = flap_area / ref_wing.areas.reference
    L              = ref_fuselage.lengths.total / Units.ft
    Bw             = ref_wing.spans.projected / Units.ft
    DG             = vehicle.mass_properties.max_takeoff / Units.lbs
    Scs            = flap_ratio * vehicle.reference_area / Units.ft**2
    design_mach    = vehicle.flight_envelope.design_mach_number
    num_pax        = vehicle.passengers 
    NENG = 0 
    for network in  vehicle.networks:
        for propulsor in network.propulsors:
            if isinstance(propulsor, RCAIDE.Library.Components.Propulsors.Turbofan) or  isinstance(propulsor, RCAIDE.Library.Components.Propulsors.Turbojet):
                NENG += 1  
    fuse_w         = ref_fuselage.width / Units.ft
    fuse_h         = ref_fuselage.heights.maximum / Units.ft   
    cargo_weight   = vehicle.payload.cargo.mass_properties.mass / Units.lbs
    
    if vehicle.passengers >= 150:
        flight_crew = 3 # number of flight crew
    else:
        flight_crew = 2
    Ns      = 4  # Number of flight control systems (typically 4)
    Kr      = 1  # assuming not a reciprocating engine
    Ktp     = 1  # assuming not a turboprop
    Nf      = 7  # number of functions performed by controls (typically 4-7)
    Rkva    = 60  # system electrical rating
    Wuav    = 1400  # uninstalled avionics weight

    WSC = 36.28 * design_mach**0.003 * Scs**0.489 * Ns**0.484 * flight_crew**0.124

    if num_pax >= 6.:
        apu_wt = 7.0 * num_pax
    else:
        apu_wt = 0.0  # no apu if less than 9 seats
    WAPU    = max(apu_wt, 70./Units.lbs) 
    WIN     = 4.509 * Kr * Ktp * flight_crew ** 0.541 * NENG * (L + Bw) ** 0.5
    WHYD    = 0.2673 * Nf * (L + Bw) ** 0.937
    WELEC   = 7.291 * Rkva ** 0.782 * (2*L) ** 0.346 * NENG ** 0.1
    WAVONC  = 1.73 * Wuav ** 0.983

    D       = (fuse_w + fuse_h) / 2.
    Sf      = np.pi * (L / D - 1.7) * D ** 2  # Fuselage wetted area, ft**2
    WFURN   = 0.0577 * flight_crew ** 0.1 * (cargo_weight) ** 0.393 * Sf ** 0.75 + 46 * num_pax
    WFURN  += 75 * flight_crew
    WFURN  += 2.5 * num_pax**1.33

    Vpr = D ** 2 * np.pi / 4 * L
    WAC = 62.36 * num_pax ** 0.25 * (Vpr / 1000) ** 0.604 * Wuav ** 0.1

    WAI = 0.002 * DG

    output                     = Data()
    output.W_flight_control    = WSC * Units.lbs
    output.W_apu               = WAPU * Units.lbs
    output.W_hyd_pnu           = WHYD * Units.lbs
    output.W_instruments       = WIN * Units.lbs
    output.W_avionics          = WAVONC * Units.lbs
    output.W_electrical        = WELEC * Units.lbs
    output.W_ac                = WAC * Units.lbs
    output.W_furnish           = WFURN * Units.lbs
    output.W_anti_ice          = WAI * Units.lbs
    output.W_systems           = WSC + WAPU + WIN + WHYD + WELEC + WAVONC + WFURN + WAC + WAI
    return output



# RCAIDE/Library/Missions/Common/Update/aerodynamics.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  Update Aerodynamics
# ----------------------------------------------------------------------------------------------------------------------
def aerodynamics(segment):
    """ Gets aerodynamics conditions
    
        Assumptions:
        +X out nose
        +Y out starboard wing
        +Z down

        Inputs:
            segment.analyses.aerodynamics_model                    [Function]
            aerodynamics_model.settings.maximum_lift_coefficient   [unitless]
            aerodynamics_model.vehicle.reference_area             [meter^2]
            segment.state.conditions.freestream.dynamic_pressure   [pascals]

        Outputs:
            conditions.aerodynamics.coefficients.lift.total [unitless]
            conditions.aerodynamics.coefficients.drag.total [unitless]
            conditions.frames.wind.force_vector [newtons]
            conditions.frames.wind.drag_force_vector [newtons]

        Properties Used:
        N/A
    """
    
    # unpack
    conditions         = segment.state.conditions
    aerodynamics_model = segment.analyses.aerodynamics
    q                  = segment.state.conditions.freestream.dynamic_pressure
    Sref               = aerodynamics_model.vehicle.reference_area
    CLmax              = aerodynamics_model.settings.maximum_lift_coefficient 
    MAC                = aerodynamics_model.vehicle.wings.main_wing.chords.mean_aerodynamic
    span               = aerodynamics_model.vehicle.wings.main_wing.spans.projected 
    
    # call aerodynamics model
    _ = aerodynamics_model(segment)     

    # Forces 
    CL = conditions.aerodynamics.coefficients.lift.total
    CD = conditions.aerodynamics.coefficients.drag.total
    CY = conditions.static_stability.coefficients.Y

    CL[q<=0.0] = 0.0
    CD[q<=0.0] = 0.0
    CL[CL>CLmax] = CLmax
    CL[CL< -CLmax] = -CLmax

    # dimensionalize
    F      = segment.state.ones_row(3) * 0.0
    F[:,2] = ( -CL * q * Sref )[:,0]
    F[:,1] = ( -CY * q * Sref )[:,0]
    F[:,0] = ( -CD * q * Sref )[:,0]

    # rewrite aerodynamic CL and CD
    conditions.aerodynamics.coefficients.lift.total  = CL
    conditions.aerodynamics.coefficients.drag.total  = CD
    conditions.frames.wind.force_vector[:,:]   = F[:,:]

    # -----------------------------------------------------------------
    # Moments
    # -----------------------------------------------------------------
    C_M = conditions.static_stability.coefficients.M
    C_L = conditions.static_stability.coefficients.L
    C_N = conditions.static_stability.coefficients.N

    C_M[q<=0.0] = 0.0

    # dimensionalize
    M      = segment.state.ones_row(3) * 0.0
    M[:,0] = (C_L[:,0] * q[:,0] * Sref * span)
    M[:,1] = (C_M[:,0] * q[:,0] * Sref * MAC)
    M[:,2] = (C_N[:,0] * q[:,0] * Sref * span)

    # pack conditions
    conditions.frames.wind.moment_vector[:,:] = M[:,:] 

    return


# RCAIDE/Methods/Energy/Propulsors/Rotor_Design/procedure_setup.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

# RCAIDE Imports 
import RCAIDE 
from RCAIDE.Framework.Core                                                        import Units  
from RCAIDE.Library.Methods.Noise.Frequency_Domain_Buildup.Rotor                  import compute_rotor_noise 
from RCAIDE.Framework.Analyses.Process                                            import Process   
from RCAIDE.Framework.Mission.Common                                              import Conditions
from RCAIDE.Library.Methods.Propulsors.Converters.Rotor.compute_rotor_performance import compute_rotor_performance 

# Python package imports   
import numpy as np 
import scipy as sp  
    
# ----------------------------------------------------------------------------------------------------------------------  
#  Procedure Setup 
# ----------------------------------------------------------------------------------------------------------------------    
def procedure_setup(): 
    
    # size the base config
    procedure = Process()
    
    # mofify blade geometry 
    procedure.modify_rotor = modify_blade_geometry
    
    # Run the rotor in hover
    procedure.hover       = run_rotor_hover
    
    # Run the rotor in the oei condition
    procedure.oei         = run_rotor_OEI 

     # Run the rotor in cruise
    procedure.ruise       = run_rotor_cruise

    # post process the results
    procedure.post_process = post_process
        
    return procedure


# ----------------------------------------------------------------------
# Update blade geometry 
# ---------------------------------------------------------------------- 
def modify_blade_geometry(nexus): 
    """ Modifies geometry of prop-rotor blade 
          
          Inputs:  
             nexus     - RCAIDE optmization framework with prop-rotor blade data structure [None]
              
          Outputs:   
             procedure - optimization methodology                                         [None]
              
          Assumptions: 
             N/A 
        
          Source:
             None
    """        
 
    # Pull out the vehicles

    vehicle_hover     = nexus.vehicle_configurations.hover
    rotor_hover       = vehicle_hover.networks.electric.propulsors.electric_rotor.rotor
    vehicle_oei       = nexus.vehicle_configurations.oei
    rotor_oei         = vehicle_oei.networks.electric.propulsors.electric_rotor.rotor    
    
    if nexus.prop_rotor_flag:  
        vehicle_cruise    = nexus.vehicle_configurations.cruise 
        rotor_cruise      = vehicle_cruise.networks.electric.propulsors.electric_rotor.rotor 
        
    airfoils = rotor_hover.airfoils      
    a_loc    = rotor_hover.airfoil_polar_stations   
        
    # Update geometry of blade
    R       = rotor_hover.tip_radius     
    B       = rotor_hover.number_of_blades 
    r       = rotor_hover.radius_distribution
    c       = updated_blade_geometry(rotor_hover.radius_distribution/rotor_hover.tip_radius ,rotor_hover.chord_r,rotor_hover.chord_p,rotor_hover.chord_q,rotor_hover.chord_t)     
    beta    = updated_blade_geometry(rotor_hover.radius_distribution/rotor_hover.tip_radius ,rotor_hover.twist_r,rotor_hover.twist_p,rotor_hover.twist_q,rotor_hover.twist_t)   
    
    # compute max thickness distribution   
    blade_area    = sp.integrate.cumulative_trapezoid(B*c, r-r[0])
    sigma         = blade_area[-1]/(np.pi*R**2)      
    t_max         = np.zeros(len(c))    
    t_c           = np.zeros(len(c))       
    t_max  = np.zeros(len(c))     
    if len(airfoils.keys())>0:
        for j,airfoil in enumerate(airfoils): 
            a_geo         = airfoil.geometry
            locs          = np.where(np.array(a_loc) == j )
            t_max[locs]   = a_geo.max_thickness*c[locs]   
     
    rotor_hover.chord_distribution          = c
    rotor_hover.twist_distribution          = beta  
    rotor_hover.mid_chord_alignment         = c/4. - c[0]/4.
    rotor_hover.max_thickness_distribution  = t_max 
    rotor_hover.thickness_to_chord          = t_c
    rotor_hover.blade_solidity              = sigma     
    vehicle_hover.store_diff() 
      

    rotor_oei.chord_distribution         = rotor_hover.chord_distribution
    rotor_oei.twist_distribution         = rotor_hover.twist_distribution
    rotor_oei.mid_chord_alignment        = rotor_hover.mid_chord_alignment  
    rotor_oei.max_thickness_distribution = rotor_hover.max_thickness_distribution
    rotor_oei.thickness_to_chord         = rotor_hover.thickness_to_chord 
    rotor_oei.blade_solidity             = rotor_hover.blade_solidity    
    vehicle_oei.store_diff()     
     
    if nexus.prop_rotor_flag: 
        rotor_cruise.chord_distribution         = rotor_hover.chord_distribution
        rotor_cruise.twist_distribution         = rotor_hover.twist_distribution
        rotor_cruise.mid_chord_alignment        = rotor_hover.mid_chord_alignment  
        rotor_cruise.max_thickness_distribution = rotor_hover.max_thickness_distribution
        rotor_cruise.thickness_to_chord         = rotor_hover.thickness_to_chord 
        rotor_cruise.blade_solidity             = rotor_hover.blade_solidity     
        vehicle_cruise.store_diff()  
    
    return nexus    


# ----------------------------------------------------------------------
#   Update blade geometry 
# ---------------------------------------------------------------------- 
def updated_blade_geometry(chi,c_r,p,q,c_t):
    """ Computes planform function of twist and chord distributron using hyperparameters  
          
          Inputs:  
             chi - prop-rotor radius distribution [None]
             c_r - hyperparameter no. 1           [None]
             p   - hyperparameter no. 2           [None]
             q   - hyperparameter no. 3           [None]
             c_t - hyperparameter no. 4           [None] 
                   
          Outputs:       
             x_lin  - function distribution       [None]
              
          Assumptions: 
             N/A 
        
          Source:
              Traub, Lance W., et al. "Effect of taper ratio at low reynolds number."
              Journal of Aircraft 52.3 (2015): 734-747.
              
    """           

    n       = np.linspace(len(chi)-1,0,len(chi))          
    theta_n = n*(np.pi/2)/len(chi)              
    y_n     = chi[-1]*np.cos(theta_n)          
    eta_n   = np.abs(y_n/chi[-1])            
    x_cos   = c_r*(1 - eta_n**p)**q + c_t*eta_n  
    x_lin   = np.interp(chi,eta_n, x_cos)  
    return x_lin 


# ----------------------------------------------------------------------
#   Run the Rotor Hover
# ----------------------------------------------------------------------  
def run_rotor_hover(nexus):
     
    # Unpack    
    network               = nexus.vehicle_configurations.hover.networks.electric 
    bus                   = network.busses.bus 
    electric_rotor        = network.propulsors.electric_rotor
    rotor                 = electric_rotor.rotor
    
    # Setup Test conditions
    alpha                   = rotor.optimization_parameters.multiobjective_aeroacoustic_weight 
    speed                   = rotor.hover.design_freestream_velocity 
    altitude                = np.array([rotor.hover.design_altitude])  
    atmosphere              = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmosphere_conditions   = atmosphere.compute_values(altitude)  
  
    segment                                             = RCAIDE.Framework.Mission.Segments.Segment()  
    conditions                                          = RCAIDE.Framework.Mission.Common.Results()
    conditions.freestream.update(atmosphere_conditions)  
    conditions.frames.inertial.velocity_vector          = np.array([[0.,0.,speed]])   
    conditions.frames.body.transform_to_inertial        = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0., -1.]]]) 
    conditions.frames.wind.transform_to_inertial        = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0.,  1.]]]) 
    conditions.frames.planet.true_course                = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0.,  1.]]])  
    segment.state.conditions                            = conditions
    
    segment.state.conditions.energy[bus.tag] = Conditions()
    segment.state.conditions.noise[bus.tag]  = Conditions() 
    electric_rotor.append_operating_conditions(segment) 
    for tag, item in  electric_rotor.items(): 
        if issubclass(type(item), RCAIDE.Library.Components.Component):
            item.append_operating_conditions(segment,electric_rotor)
    
    rotor_conditions                      =  segment.state.conditions.energy[electric_rotor.tag][rotor.tag]     
    rotor_conditions.omega                = (atmosphere_conditions.speed_of_sound*rotor.hover.design_tip_mach)/rotor.tip_radius
    rotor_conditions.pitch_command[:,0]   = rotor.hover.design_pitch_command
    
    compute_rotor_performance(electric_rotor,segment.state)   
     
    # Pack the results
    nexus.results.hover.thrust                       = -segment.state.conditions.energy[electric_rotor.tag][rotor.tag].thrust[0,2]  
    nexus.results.hover.torque                       = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].torque[0][0]
    nexus.results.hover.power                        = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].power[0][0]
    nexus.results.hover.power_c                      = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].power_coefficient[0][0]
    nexus.results.hover.thurst_c                     = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].thrust_coefficient[0][0]
    nexus.results.hover.omega                        = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].omega[0][0]
    nexus.results.hover.max_sectional_cl             = np.max(segment.state.conditions.energy[electric_rotor.tag][rotor.tag].lift_coefficient[0]) 
    nexus.results.hover.mean_CL                      = np.mean(segment.state.conditions.energy[electric_rotor.tag][rotor.tag].lift_coefficient[0]) 
    nexus.results.hover.figure_of_merit              = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].figure_of_merit[0][0]  
    nexus.results.hover.efficiency                   = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].efficiency[0][0] 
    nexus.results.hover.conditions                   = conditions  
                
    # microphone locations             
    ctrl_pts                                         = 1 
    theta                                            = rotor.optimization_parameters.noise_evaluation_angle 
    S_hover                                          = np.maximum(altitude[0],20*Units.feet)  
    mic_positions_hover                              = np.array([[0.0 , S_hover*np.sin(theta)  ,S_hover*np.cos(theta)]])      
    
    # Run noise model  
    conditions.noise[bus.tag]                        = RCAIDE.Framework.Mission.Common.Conditions()      
    conditions.noise[electric_rotor.tag]             = RCAIDE.Framework.Mission.Common.Conditions()  
    conditions.noise.relative_microphone_locations   = np.repeat(mic_positions_hover[ np.newaxis,:,: ],1,axis=0)
    conditions.aerodynamics.angles.alpha             = np.ones((ctrl_pts,1))* 0. * Units.degrees 
    segment                                          = RCAIDE.Framework.Mission.Segments.Segment() 
    segment.state.conditions                         = conditions
    segment.state.conditions.expand_rows(ctrl_pts)  
    noise                                            = RCAIDE.Framework.Analyses.Noise.Frequency_Domain_Buildup() 
    settings                                         = noise.settings   
    num_mic                                          = len(conditions.noise.relative_microphone_locations[0])  
    conditions.noise.number_of_microphones           = num_mic   
    
    if alpha != 1: 
        compute_rotor_noise(bus,electric_rotor,segment,settings)    
        nexus.results.hover.mean_SPL   = np.mean(conditions.noise[electric_rotor.tag][rotor.tag].SPL_dBA) 
    else: 
        nexus.results.hover.mean_SPL   = 0  

    return nexus

# ----------------------------------------------------------------------
#   One Engine Inoperative 
# ---------------------------------------------------------------------- 
def run_rotor_OEI(nexus):
    
    # Unpack   
    network               = nexus.vehicle_configurations.oei.networks.electric 
    bus                   = network.busses.bus 
    electric_rotor        = network.propulsors.electric_rotor    
    rotor                 = electric_rotor.rotor
    
    # Setup Test conditions
    speed                 = rotor.oei.design_freestream_velocity 
    altitude              = np.array([rotor.oei.design_altitude]) 
    atmosphere            = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmosphere_conditions = atmosphere.compute_values(altitude)  
  
  
    segment                                             = RCAIDE.Framework.Mission.Segments.Segment()  
    conditions                                          = RCAIDE.Framework.Mission.Common.Results()
    conditions.freestream.update(atmosphere_conditions)  
    conditions.frames.inertial.velocity_vector          = np.array([[0.,0.,speed]])   
    conditions.frames.body.transform_to_inertial        = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0., -1.]]]) 
    conditions.frames.wind.transform_to_inertial        = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0.,  1.]]]) 
    conditions.frames.planet.true_course                = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0.,  1.]]]) 
    segment.state.conditions                            = conditions
    
    segment.state.conditions.energy[bus.tag] = Conditions()
    segment.state.conditions.noise[bus.tag]  = Conditions()
    electric_rotor.append_operating_conditions(segment) 
    for tag, item in  electric_rotor.items(): 
        if issubclass(type(item), RCAIDE.Library.Components.Component):
            item.append_operating_conditions(segment,electric_rotor) 
                
    rotor_conditions                      =  segment.state.conditions.energy[electric_rotor.tag][rotor.tag]     
    rotor_conditions.omega                = (atmosphere_conditions.speed_of_sound*rotor.oei.design_tip_mach)/rotor.tip_radius
    rotor_conditions.pitch_command[:,0]   = rotor.oei.design_pitch_command
    
    compute_rotor_performance(electric_rotor,segment.state)   
            
    # Pack the results
    nexus.results.oei.thrust       = -segment.state.conditions.energy[electric_rotor.tag][rotor.tag].thrust[0,2]  
    nexus.results.oei.torque       = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].torque[0][0]
    nexus.results.oei.power        = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].power[0][0]
    nexus.results.oei.power_c      = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].power_coefficient[0][0]
    nexus.results.oei.omega        = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].omega[0][0]
    nexus.results.oei.thurst_c     = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].thrust_coefficient[0][0]
    nexus.results.oei.efficiency   = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].efficiency[0][0]
    nexus.results.oei.conditions   = conditions 
    
    return nexus

# ----------------------------------------------------------------------
#   Run the Rotor Cruise
# ----------------------------------------------------------------------  
def run_rotor_cruise(nexus):
 
    if nexus.prop_rotor_flag:     
        network         = nexus.vehicle_configurations.cruise.networks.electric 
        bus             = network.busses.bus 
        electric_rotor  = network.propulsors.electric_rotor
        rotor           = electric_rotor.rotor 
        alpha           = rotor.optimization_parameters.multiobjective_aeroacoustic_weight       
        
        # Setup Test conditions
        speed                 = rotor.cruise.design_freestream_velocity 
        altitude              = np.array([rotor.cruise.design_altitude])  
        atmosphere            = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
        atmosphere_conditions = atmosphere.compute_values(altitude) 
      
        segment                                             = RCAIDE.Framework.Mission.Segments.Segment()  
        conditions                                          = RCAIDE.Framework.Mission.Common.Results()
        conditions.freestream.update(atmosphere_conditions)  
        conditions.frames.inertial.velocity_vector          = np.array([[0.,0.,speed]])   
        conditions.frames.body.transform_to_inertial        = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0., -1.]]]) 
        conditions.frames.wind.transform_to_inertial        = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0.,  1.]]]) 
        conditions.frames.planet.true_course                = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0.,  1.]]]) 
        segment.state.conditions                            = conditions
        
        segment.state.conditions.energy[bus.tag] = Conditions()
        segment.state.conditions.noise[bus.tag]  = Conditions()
        electric_rotor.append_operating_conditions(segment) 
        for tag, item in  electric_rotor.items(): 
            if issubclass(type(item), RCAIDE.Library.Components.Component):
                item.append_operating_conditions(segment,electric_rotor) 
            
        rotor_conditions                      =  segment.state.conditions.energy[electric_rotor.tag][rotor.tag]     
        rotor_conditions.omega                = (atmosphere_conditions.speed_of_sound*rotor.cruise.design_tip_mach)/rotor.tip_radius
        rotor_conditions.pitch_command[:,0]   = rotor.cruise.design_pitch_command
        
        compute_rotor_performance(electric_rotor,segment.state)   
        
        # Pack the results
        nexus.results.cruise.thrust                      = -segment.state.conditions.energy[electric_rotor.tag][rotor.tag].thrust[0,2] 
        nexus.results.cruise.torque                      = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].torque[0][0]
        nexus.results.cruise.power                       = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].power[0][0]
        nexus.results.cruise.power_c                     = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].power_coefficient[0][0]
        nexus.results.cruise.omega                       = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].omega[0][0]
        nexus.results.cruise.thurst_c                    = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].thrust_coefficient[0][0]
        nexus.results.cruise.max_sectional_cl            = np.max(segment.state.conditions.energy[electric_rotor.tag][rotor.tag].lift_coefficient[0]) 
        nexus.results.cruise.mean_CL                     = np.mean(segment.state.conditions.energy[electric_rotor.tag][rotor.tag].lift_coefficient[0])  
        nexus.results.cruise.efficiency                  = segment.state.conditions.energy[electric_rotor.tag][rotor.tag].efficiency[0][0]
        nexus.results.cruise.conditions                  = conditions  
                    
        # microphone locations            
        ctrl_pts                                         = 1 
        theta                                            = rotor.optimization_parameters.noise_evaluation_angle 
        S_cruise                                         = np.maximum(altitude[0],20*Units.feet)  
        mic_positions_cruise                             = np.array([[0.0 ,S_cruise*np.sin(theta)  ,S_cruise*np.cos(theta)]])      
        
        # Run noise model  
        conditions.noise[bus.tag]                        = RCAIDE.Framework.Mission.Common.Conditions()      
        conditions.noise[electric_rotor.tag]    = RCAIDE.Framework.Mission.Common.Conditions()  
        conditions.noise.relative_microphone_locations   = np.repeat(mic_positions_cruise[ np.newaxis,:,: ],1,axis=0)
        conditions.aerodynamics.angles.alpha             = np.ones((ctrl_pts,1))* 0. * Units.degrees 
        segment                                          = RCAIDE.Framework.Mission.Segments.Segment() 
        segment.state.conditions                         = conditions
        segment.state.conditions.expand_rows(ctrl_pts)  
        noise                                            = RCAIDE.Framework.Analyses.Noise.Frequency_Domain_Buildup() 
        settings                                         = noise.settings   
        num_mic                                          = len(conditions.noise.relative_microphone_locations[0])  
        conditions.noise.number_of_microphones           = num_mic    
        
        if alpha != 1: 
            compute_rotor_noise(bus,electric_rotor,segment,settings)      
            nexus.results.cruise.mean_SPL   = np.mean(conditions.noise[electric_rotor.tag][rotor.tag].SPL_dBA)   
        else:
            nexus.results.cruise.mean_SPL   = 0  
            
    else:     
        nexus.results.cruise.thrust           = 0.0
        nexus.results.cruise.torque           = 0.0
        nexus.results.cruise.power            = 0.0
        nexus.results.cruise.power_c          = 0.0
        nexus.results.cruise.thurst_c         = 0.0
        nexus.results.cruise.omega            = 0.0
        nexus.results.cruise.max_sectional_cl = 0.0
        nexus.results.cruise.mean_CL          = 0.0
        nexus.results.cruise.efficiency       = 0.0  
        nexus.results.cruise.mean_SPL         = 0.0
        nexus.results.cruise.noise_data       = None   

    return nexus
   

# ----------------------------------------------------------------------
#   Post Process Results to give back to the optimizer
# ----------------------------------------------------------------------   
def post_process(nexus):
     
    rotor                           = nexus.vehicle_configurations.hover.networks.electric.propulsors.electric_rotor.rotor  
    rotor_oei                       = nexus.vehicle_configurations.oei.networks.electric.propulsors.electric_rotor.rotor
    alpha                           = rotor.optimization_parameters.multiobjective_aeroacoustic_weight
    beta                            = rotor.optimization_parameters.multiobjective_performance_weight
    gamma                           = rotor.optimization_parameters.multiobjective_acoustic_weight
    ideal_SPL                       = rotor.optimization_parameters.ideal_SPL_dBA  
    ideal_efficiency                = rotor.optimization_parameters.ideal_efficiency      
    ideal_FoM                       = rotor.optimization_parameters.ideal_figure_of_merit  
    print_iter                      = nexus.print_iterations  
    mean_CL_hover                   = nexus.results.hover.mean_CL
    omega_hover                     = nexus.results.hover.omega
    FM_hover                        = nexus.results.hover.figure_of_merit  
    
    # q to p ratios 
    summary                                 = nexus.summary 
    summary.max_sectional_cl_hover          = nexus.results.hover.max_sectional_cl
    summary.chord_p_to_q_ratio              = rotor.chord_p/rotor.chord_q
    summary.twist_p_to_q_ratio              = rotor.twist_p/rotor.twist_q      
    summary.blade_taper_constraint_1        = rotor.chord_distribution[-1]/rotor.chord_distribution[0]
    summary.blade_taper_constraint_2        = rotor.chord_distribution[-1]/rotor.chord_distribution[0]
    summary.blade_twist_constraint          = rotor.twist_distribution [0] - rotor.twist_distribution [-1] 
    summary.OEI_hover_thrust_power_residual = abs(nexus.results.oei.thrust - rotor.oei.design_thrust) / nexus.results.oei.thrust
            
    # thrust/power residuals  
    if rotor.hover.design_thrust == None:
        summary.hover_thrust_power_residual = abs(nexus.results.hover.power - rotor.hover.design_power) / rotor.hover.design_power
    else: 
        summary.hover_thrust_power_residual = abs(nexus.results.hover.thrust - rotor.hover.design_thrust) /rotor.hover.design_thrust 

    # oei
    if rotor.oei.design_thrust == None:
        summary.oei_thrust_power_residual =  abs(nexus.results.oei.power - rotor.oei.design_power) / rotor.oei.design_power
    else: 
        summary.oei_thrust_power_residual = abs(nexus.results.oei.thrust - rotor.oei.design_thrust) /rotor.oei.design_thrust 
    
        
    if nexus.prop_rotor_flag: 
        if rotor.cruise.design_thrust == None:
            summary.cruise_thrust_power_residual = abs(nexus.results.cruise.power - rotor.cruise.design_power) /rotor.cruise.design_power 
        else: 
            summary.cruise_thrust_power_residual =  abs(nexus.results.cruise.thrust - rotor.cruise.design_thrust) /rotor.cruise.design_thrust    
            
    # -------------------------------------------------------
    # OBJECTIVE FUNCTION
    # -------------------------------------------------------   
    performance_objective  = ((ideal_FoM - FM_hover)/ideal_FoM)*beta +  ((ideal_efficiency - nexus.results.cruise.efficiency)/ideal_efficiency)*(1-beta) 
    
    acoustic_objective     = ((nexus.results.hover.mean_SPL  - ideal_SPL)/ideal_SPL)*gamma  + ((nexus.results.cruise.mean_SPL - ideal_SPL)/ideal_SPL)*(1-gamma) 
 
    summary.objective      = (performance_objective*alpha + acoustic_objective*(1-alpha))  
    

    if nexus.prop_rotor_flag:  
        rotor_cru  = nexus.vehicle_configurations.cruise.networks.electric.propulsors.electric_rotor.rotor         
        summary.max_sectional_cl_cruise = nexus.results.cruise.max_sectional_cl   
        
    # -------------------------------------------------------
    # PRINT ITERATION PERFOMRMANCE
    # -------------------------------------------------------      
    if print_iter:
        print("Aeroacoustic Weight          : " + str(alpha))  
        print("Multiobj. Performance Weight : " + str(beta))  
        print("Multiobj. Acoustic Weight    : " + str(gamma)) 
        print("Performance Obj              : " + str(performance_objective))   
        print("Acoustic Obj                 : " + str(acoustic_objective))  
        print("Aeroacoustic Obj             : " + str(summary.objective))    
        print("Blade Taper                  : " + str(summary.blade_taper_constraint_1))
        print("Hover RPM                    : " + str(omega_hover/Units.rpm))     
        if rotor.hover.design_thrust == None: 
            print("Hover Power                  : " + str(nexus.results.hover.power))  
        if rotor.hover.design_power == None: 
            print("Hover Thrust                 : " + str(nexus.results.hover.thrust))  
        print("Hover Average SPL            : " + str(nexus.results.hover.mean_SPL))    
        print("Hover Tip Mach               : " + str(rotor.hover.design_tip_mach)) 
        print("Hover Thrust/Power Residual  : " + str(summary.hover_thrust_power_residual)) 
        print("Hover Figure of Merit        : " + str(FM_hover))  
        print("Hover Max Sectional Cl       : " + str(summary.max_sectional_cl_hover)) 
        print("Hover Blade CL               : " + str(mean_CL_hover))    
        print("OEI Thrust                   : " + str(nexus.results.oei.thrust)) 
        print("OEI Thrust/Power Residual    : " + str(summary.OEI_hover_thrust_power_residual)) 
        print("OEI Tip Mach                 : " + str(rotor_oei.oei.design_tip_mach))  
        print("OEI Collective (deg)         : " + str(rotor_oei.design_pitch_command/Units.degrees)) 
        if nexus.prop_rotor_flag:    
            print("Cruise RPM                   : " + str(nexus.results.cruise.omega/Units.rpm))    
            print("Cruise Collective (deg)      : " + str(rotor_cru.design_pitch_command/Units.degrees)) 
            if rotor_cru.cruise.design_thrust == None:  
                print("Cruise Power                 : " + str(nexus.results.cruise.power)) 
            if rotor_cru.cruise.design_power == None:  
                print("Cruise Thrust                : " + str(nexus.results.cruise.thrust))   
            print("Cruise Tip Mach              : " + str(rotor_cru.cruise.design_tip_mach))  
            print("Cruise Thrust/Power Residual : " + str(summary.cruise_thrust_power_residual))
            print("Cruise Efficiency            : " + str(nexus.results.cruise.efficiency)) 
            print("Cruise Max Sectional Cl      : " + str(summary.max_sectional_cl_cruise))  
            print("Cruise Blade CL              : " + str(nexus.results.cruise.mean_CL))  
        print("\n\n") 

   
    return nexus    



# RCAIDE/Library/Components/Nacelles/__init__.py

"""
Module containing nacelle components for aircraft propulsion system integration. This module 
provides different nacelle configurations including stack nacelles and bodies of revolution 
for various engine installations.

See Also
--------
RCAIDE.Library.Components.Fuselages
    Related module for fuselage components that may interface with engine nacelles
RCAIDE.Library.Components.Landing_Gear
    Related module for landing gear design which may affect nacelle ground clearance
RCAIDE.Library.Components.Airfoils
    Related module for airfoil definitions that may influence nacelle-wing integration
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Nacelle                    import Nacelle 
from .Stack_Nacelle              import Stack_Nacelle
from .Body_of_Revolution_Nacelle import Body_of_Revolution_Nacelle

from . import Segments


# RCAIDE/Framework/Analyses/Mission/Segments/Climb/__init__.py
# 

"""RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Constant_Dynamic_Pressure_Constant_Angle import Constant_Dynamic_Pressure_Constant_Angle
from .Constant_Dynamic_Pressure_Constant_Rate  import Constant_Dynamic_Pressure_Constant_Rate
from .Constant_Mach_Constant_Angle             import Constant_Mach_Constant_Angle
from .Constant_Mach_Constant_Rate              import Constant_Mach_Constant_Rate
from .Constant_Mach_Linear_Altitude            import Constant_Mach_Linear_Altitude
from .Constant_Speed_Constant_Angle            import Constant_Speed_Constant_Angle
from .Constant_Speed_Constant_Angle_Noise      import Constant_Speed_Constant_Angle_Noise
from .Constant_Speed_Constant_Rate             import Constant_Speed_Constant_Rate
from .Constant_Speed_Linear_Altitude           import Constant_Speed_Linear_Altitude
from .Constant_Throttle_Constant_Speed         import Constant_Throttle_Constant_Speed
from .Linear_Mach_Constant_Rate                import Linear_Mach_Constant_Rate
from .Linear_Speed_Constant_Rate               import Linear_Speed_Constant_Rate
from .Constant_EAS_Constant_Rate               import Constant_EAS_Constant_Rate
from .Constant_CAS_Constant_Rate               import Constant_CAS_Constant_Rate 


# RCAIDE/Library/Components/Airfoils/__init__.py
# 

"""
Airfoils module providing classes and methods for defining and analyzing aircraft airfoil geometries

This module contains implementations of various airfoil types including general airfoils
and specific NACA series airfoils. It provides functionality for defining airfoil geometry.

Notes
-----
The airfoil classes in this module are designed to work with RCAIDE's aerodynamic analysis
tools and can be integrated into complete aircraft configurations. Airfoil data can be
imported from external sources or generated using geometric definitions.

See Also
--------
RCAIDE.Library.Components.Wings
    Wing components that utilize airfoil definitions
RCAIDE.Framework.Analyses.Aerodynamics
    Aerodynamic analysis tools that work with airfoil components
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Airfoil               import Airfoil
from .NACA_4_Series_Airfoil import NACA_4_Series_Airfoil
from .Biconvex_Airfoil      import Biconvex_Airfoil


# RCAIDE/Methods/Aerodynamics/Common/Gas_Dymamics/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

from .fm_id                  import fm_id
from .fm_solver              import fm_solver
from .oblique_shock          import oblique_shock_relations, theta_beta_mach
from .rayleigh               import rayleigh
from .nozzle_calculations    import exit_Mach_shock, mach_area, normal_shock, pressure_ratio_isentropic, pressure_ratio_shock_in_nozzle 


# RCAIDE/Methods/Energy/Propulsors/Electric_Ducted_Fan_Propulsor/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .compute_electric_ducted_fan_performance  import compute_electric_ducted_fan_performance
from .append_electric_ducted_fan_conditions    import append_electric_ducted_fan_conditions
from .pack_electric_ducted_fan_residuals       import pack_electric_ducted_fan_residuals
from .unpack_electric_ducted_fan_unknowns      import unpack_electric_ducted_fan_unknowns


# RCAIDE/Library/Plots/Stability/plot_stability_forces_and_moments.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_flight_forces_and_moments(results,
                             save_figure = False,
                             show_legend = True,
                             save_filename = "Intertial_Forces_and_Moments",
                             file_type = ".png",
                             width = 12, height = 8):
    """This plots the aerodynamic forces 
    """ 

    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))     
  
    
    fig   = plt.figure(save_filename)
    fig.set_size_inches(width,height) 

    axis_1 = plt.subplot(3,2,1)
    axis_2 = plt.subplot(3,2,2)
    axis_3 = plt.subplot(3,2,3)
    axis_4 = plt.subplot(3,2,4)
    axis_5 = plt.subplot(3,2,5)  
    axis_6 = plt.subplot(3,2,6)    
    
    for i in range(len(results.segments)): 
        time   = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min 
        X = results.segments[i].conditions.frames.inertial.total_force_vector[:,0]
        Y = results.segments[i].conditions.frames.inertial.total_force_vector[:,1]
        Z = results.segments[i].conditions.frames.inertial.total_force_vector[:,2]
        L = results.segments[i].conditions.frames.inertial.total_moment_vector[:,0]
        M = results.segments[i].conditions.frames.inertial.total_moment_vector[:,1]
        N = results.segments[i].conditions.frames.inertial.total_moment_vector[:,2]
        
                       
        segment_tag  =  results.segments[i].tag
        segment_name = segment_tag.replace('_', ' ') 
        axis_1.plot(time,X, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width, label = segment_name) 
        axis_1.set_ylabel(r'X Moment (N)')
        axis_1.set_ylim([-1000, 1000]) 
        set_axes(axis_1)                
        
        axis_2.set_ylabel(r'Roll Moment (Nm)')
        axis_2.plot(time,L, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width) 
        axis_2.set_ylim([-1000, 1000]) 
        set_axes(axis_2)     
        
        axis_3.plot(time,Y, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width) 
        axis_3.set_ylabel(r'Y force (N)')
        axis_3.set_ylim([-1000, 1000]) 
        set_axes(axis_3)  
        
        axis_4.plot(time,M, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width) 
        axis_4.set_ylabel(r'Pitch Moment (Nm)')
        axis_4.set_ylim([-1000, 1000]) 
        set_axes(axis_4) 

        axis_5.plot(time, Z, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
        axis_5.set_xlabel('Time (mins)')
        axis_5.set_ylabel(r'Z Force (m)')
        axis_5.set_ylim([-1000, 1000]) 
        set_axes(axis_5)    

        axis_6.plot(time, N, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
        axis_6.set_xlabel('Time (mins)')
        axis_6.set_ylabel(r'Yaw Moment (Nm)')
        axis_6.set_ylim([-1000, 1000]) 
        set_axes(axis_6)   
 
    if show_legend:    
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4) 
        leg.set_title('Flight Segment', prop={'size': ps.legend_font_size, 'weight': 'heavy'})    
    
    # Adjusting the sub-plots for legend 
    fig.tight_layout()
    fig.subplots_adjust(top=0.8) 
    
    # set title of plot 
    title_text    = 'Intertial Forces and Moments'      
    fig.suptitle(title_text)
    
    if save_figure:
        plt.savefig(save_filename + file_type)   
    return fig 



# RCAIDE/Library/Missions/Common/Update/thrust.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  Update Thrust
# ---------------------------------------------------------------------------------------------------------------------- 
def thrust(segment):
    """ Updates the thrust vector of the vehicle from the propulsors 
        
        Assumptions:
        N/A
        
        Inputs:
            None 
                 
        Outputs: 
            None
      
        Properties Used:
        N/A
                    
    """ 

    # unpack
    energy_model = segment.analyses.energy

    # evaluate
    energy_model.evaluate(segment.state)

    # pack conditions
    conditions = segment.state.conditions
    conditions.frames.body.thrust_force_vector       = conditions.energy.thrust_force_vector
    conditions.frames.body.thrust_moment_vector      = conditions.energy.thrust_moment_vector
    conditions.weights.vehicle_mass_rate             = conditions.energy.vehicle_mass_rate 

    if "vehicle_additional_fuel_rate" in conditions.energy: 
        conditions.weights.has_additional_fuel             = True
        conditions.weights.vehicle_fuel_rate               = conditions.energy.vehicle_fuel_rate
        conditions.weights.vehicle_additional_fuel_rate    = conditions.energy.vehicle_additional_fuel_rate  


# RCAIDE/Library/Attributes/Solids/Solid.py
#  
 
#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------

from RCAIDE.Framework.Core import Data

#-------------------------------------------------------------------------------
# Solid Data Class
#------------------------------------------------------------------------------- 
class Solid(Data):
    """ Default Template for Solid Attribute Classes  
    """

    def __defaults__(self):
        """Default Instantiation of Physical Property Values
        
        Assumptions:
            None
        
        Source:
            None
        """

        self.ultimate_tensile_strength  = None
        self.ultimate_shear_strength    = None
        self.ultimate_bearing_strength  = None
        self.yield_tensile_strength     = None
        self.yield_shear_strength       = None
        self.yield_bearing_strength     = None
        self.minimum_gage_thickness     = None
        self.density                    = None



## @ingroup Methods-Noise-Frequency_Domain_Buildup-Rotor
# RCAIDE/Methods/Noise/Frequency_Domain_Buildup/Rotor/broadband_noise.py
# 
# 
# Created:  Sep 2024, Niranjan Nanjappa  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE 
from RCAIDE.Framework.Core import Units  
from RCAIDE.Library.Methods.Noise.Frequency_Domain_Buildup.Rotor.BPM_boundary_layer_properties import BPM_boundary_layer_properties
from RCAIDE.Library.Methods.Noise.Frequency_Domain_Buildup.Rotor.LBL_VS_broadband_noise        import LBL_VS_broadband_noise
from RCAIDE.Library.Methods.Noise.Frequency_Domain_Buildup.Rotor.TBL_TE_broadband_noise        import TBL_TE_broadband_noise
from RCAIDE.Library.Methods.Noise.Frequency_Domain_Buildup.Rotor.TIP_broadband_noise           import TIP_broadband_noise 
from RCAIDE.Library.Methods.Noise.Frequency_Domain_Buildup.Rotor.noise_directivities           import noise_directivities

# Python Package imports 
import numpy as np  
 
# ----------------------------------------------------------------------------------------------------------------------
# Compute Broadband Noise 
# ----------------------------------------------------------------------------------------------------------------------
## @ingroup Methods-Noise-Frequency_Domain_Buildup-Rotor
def broadband_noise(conditions,propulsor_conditions,coordinates,rotor,settings,Noise,cpt):
    '''This computes the trailing edge noise compoment of broadband noise of a propeller or 
    lift-rotor in the frequency domain. Boundary layer properties are computed using RCAIDE's 
    panel method.
    
    Assumptions:
        Boundary layer thickness (delta) appear to be an order of magnitude off at the trailing edge and 
        correction factor of 0.1 is used. See lines 255 and 256 
        
    Source: 
        Li, Sicheng Kevin, and Seongkyu Lee. "Prediction of Urban Air Mobility Multirotor VTOL Broadband Noise
        Using UCD-QuietFly." Journal of the American Helicopter Society (2021).
    
    Inputs:  
        freestream                                   - freestream data structure                                                          [m/s]
        angle_of_attack                              - aircraft angle of attack                                                           [rad]
        bspv                                         - rotor blade section trailing position edge vectors                                 [m]
        velocity_vector                              - velocity vector of aircraft                                                        [m/s]
        rotors                                       - data structure of rotors                                                           [None] 
        aeroacoustic_data                            - data structure of acoustic data                                                    [None] 
        settings                                     - accoustic settings                                                                 [None] 
        res                                          - results data structure                                                             [None] 
    
    Outputs 
       res.                                           *acoustic data is stored and passed in data structures*                                          
           SPL_prop_broadband_spectrum               - broadband noise in blade passing frequency spectrum                                [dB]
           SPL_prop_broadband_spectrum_dBA           - dBA-Weighted broadband noise in blade passing frequency spectrum                   [dbA]     
           SPL_prop_broadband_1_3_spectrum           - broadband noise in 1/3 octave spectrum                                             [dB]
           SPL_prop_broadband_1_3_spectrum_dBA       - dBA-Weighted broadband noise in 1/3 octave spectrum                                [dBA]                               
           p_pref_azimuthal_broadband                - azimuthal varying pressure ratio of broadband noise                                [Unitless]       
           p_pref_azimuthal_broadband_dBA            - azimuthal varying pressure ratio of dBA-weighted broadband noise                   [Unitless]     
           SPL_prop_azimuthal_broadband_spectrum     - azimuthal varying broadband noise in blade passing frequency spectrum              [dB]      
           SPL_prop_azimuthal_broadband_spectrum_dBA - azimuthal varying dBA-Weighted broadband noise in blade passing frequency spectrum [dbA]   
        
    Properties Used:
        N/A   
    '''     
    aeroacoustic_data       = propulsor_conditions[rotor.tag]
    num_cpt                 = 1
    num_mic                 = len(coordinates.X[0,:,0,0,0])  
    num_blades              = len(coordinates.X[0,0,:,0,0])
    num_sec                 = len(coordinates.X[0,0,0,:,0]) 
    frequency               = settings.center_frequencies
    num_cf                  = len(frequency)
    num_az                  = aeroacoustic_data.number_azimuthal_stations
    
    # ----------------------------------------------------------------------------------
    # Trailing Edge Noise
    # ---------------------------------------------------------------------------------- 
    freestream         = conditions.freestream
    speed_of_sound     = freestream.speed_of_sound          # speed of sound
    density            = freestream.density                 # air density 
    dyna_visc          = freestream.dynamic_viscosity  
    alpha              = aeroacoustic_data.disc_effective_angle_of_attack/Units.degrees 
    alpha_tip          = aeroacoustic_data.disc_effective_angle_of_attack[:,-1]/Units.degrees  
    disc_Re            = aeroacoustic_data.disc_reynolds_number
    disc_speed         = aeroacoustic_data.disc_velocity
    disc_Ma            = aeroacoustic_data.disc_Mach_number
    Vt                 = aeroacoustic_data.disc_tangential_velocity  
    Va                 = aeroacoustic_data.disc_axial_velocity                
    blade_chords       = rotor.chord_distribution           # blade chord    
    r                  = rotor.radius_distribution          # radial location   
    Omega              = aeroacoustic_data.omega            # angular velocity    
    L                  = np.zeros_like(r)
    del_r              = r[1:] - r[:-1]
    L[0]               = del_r[0]
    L[-1]              = del_r[-1]
    L[1:-1]            = (del_r[:-1]+ del_r[1:])/2

    if np.all(Omega == 0):
        Noise.p_pref_broadband                          = np.zeros((num_cpt,num_mic,num_cf)) 
        Noise.SPL_prop_broadband_spectrum               = np.zeros_like(Noise.p_pref_broadband)
        Noise.SPL_prop_broadband_spectrum_dBA           = np.zeros_like(Noise.p_pref_broadband)
        Noise.SPL_prop_broadband_1_3_spectrum           = np.zeros((num_cpt,num_mic,num_cf))
        Noise.SPL_prop_broadband_1_3_spectrum_dBA       = np.zeros((num_cpt,num_mic,num_cf))
        Noise.p_pref_azimuthal_broadband                = np.zeros((num_cpt,num_mic,num_cf))
        Noise.p_pref_azimuthal_broadband_dBA            = np.zeros_like(Noise.p_pref_azimuthal_broadband)
        Noise.SPL_prop_azimuthal_broadband_spectrum     = np.zeros_like(Noise.p_pref_azimuthal_broadband)
        Noise.SPL_prop_azimuthal_broadband_spectrum_dBA = np.zeros_like(Noise.p_pref_azimuthal_broadband)
    else: 
        # dimension of matrices [control pt, microphone, # blades, # blade sections, # center frequencies, # azimuthal stations] 
        c                 = np.tile(blade_chords[None,None,None,:,None,None],(num_cpt,num_mic,num_blades,1,num_cf,num_az))
        L                 = np.tile(L[None,None,None,:,None,None],(num_cpt,num_mic,num_blades,1,num_cf,num_az))
        f                 = np.tile(frequency[None,None,None,None,:,None],(num_cpt,num_mic,num_blades,num_sec,1,num_az)) 
        
        alpha_disk        = np.tile(alpha[cpt,None,None,:,None,:],(1,num_mic,num_blades,1,num_cf,1))
        V                 = np.zeros((num_cpt,num_mic,num_blades,num_sec,num_cf,num_az,3))
        V[:,:,:,:,:,:,0]  = -np.tile(Vt[cpt,None,None,:,None,:],(1,num_mic,num_blades,1,num_cf,1)) 
        V[:,:,:,:,:,:,2]  = np.tile(Va[cpt,None,None,:,None,:],(1,num_mic,num_blades,1,num_cf,1))  
        V_tot             = np.linalg.norm(V, axis=6)
        alpha_tip         = np.tile(alpha_tip[cpt,None,None,None,None,:],(1,num_mic,num_blades,num_sec,num_cf,1))  
        c_0               = np.tile(speed_of_sound[cpt,:,None,None,None,None],(1,num_mic,num_blades,num_sec,num_cf,num_az))
        rho               = np.tile(density[cpt,:,None,None,None,None],(1,num_mic,num_blades,num_sec,num_cf,num_az)) 
        mu                = np.tile(dyna_visc[cpt,:,None,None,None,None],(1,num_mic,num_blades,num_sec,num_cf,num_az)) 
        R_c               = np.tile(disc_Re[cpt,None,None,:,None,:],(1,num_mic,num_blades,1,num_cf,1))
        U                 = np.tile(disc_speed[cpt,None,None,:,None,:],(1,num_mic,num_blades,1,num_cf,1))
        M                 = np.tile(disc_Ma[cpt,None,None,:,None,:],(1,num_mic,num_blades,1,num_cf,1)) # U/c_0 
        M_tot             = V_tot/c_0   
          
        X_prime_r         = np.tile(coordinates.X_prime_r[cpt,:,:,:,None,None,:],(1,1,1,1,num_cf,num_az,1))
        cos_zeta_r        = np.sum(X_prime_r*V, axis = 6)/(np.linalg.norm(X_prime_r, axis = 6)*V_tot) 
        r_er              = np.tile(np.linalg.norm(coordinates.X_e_r[cpt], axis = 3)[None,:,:,:,None,None],(1,1,1,1,num_cf,num_az))           
        Phi_er            = np.tile(coordinates.phi_e_r[cpt,:,:,:,None,None],(1,1,1,1,num_cf,num_az))
        Theta_er          = np.tile(coordinates.theta_e_r[cpt,:,:,:,None,None],(1,1,1,1,num_cf,num_az))    
        
        # flatten matrices 
        R_c        = flatten_matrix(R_c,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        c          = flatten_matrix(c,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        alpha_star = flatten_matrix(alpha_disk,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        alpha_tip  = flatten_matrix(alpha_tip,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        U          = flatten_matrix(U,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        cos_zeta_r = flatten_matrix(cos_zeta_r,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        M_tot      = flatten_matrix(M_tot,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        f          = flatten_matrix(f,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        c_0        = flatten_matrix(c_0,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        rho        = flatten_matrix(rho,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        r_er       = flatten_matrix(r_er,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        mu         = flatten_matrix(mu,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        L          = flatten_matrix(L,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        M          = flatten_matrix(M,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        Phi_er     = flatten_matrix(Phi_er,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        Theta_er   = flatten_matrix(Theta_er,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        
        
        # calculation of boundary layer properties, eqns 2 - 16   
        # boundary layer properies of tripped and untripped at 0 angle of attack   
        boundary_layer_data  = BPM_boundary_layer_properties(R_c,c,alpha_star)  
    
        # define simulation variables/constants   
        Re_delta_star_p_untripped = boundary_layer_data.delta_star_p_untripped*U*rho/mu
        Re_delta_star_p_tripped   = boundary_layer_data.delta_star_p_tripped*U*rho/mu  
    
        # calculation of directivitiy terms , eqns 24 - 50 
        Dbar_h, Dbar_l = noise_directivities(Theta_er,Phi_er,cos_zeta_r,M_tot) 
          
        # calculation of turbulent boundary layer - trailing edge noise,  eqns 24 - 50 
        SPL_TBL_TE_tripped   = TBL_TE_broadband_noise(f,r_er,L,U,M,R_c,Dbar_h,Dbar_l,Re_delta_star_p_tripped,
                                                  boundary_layer_data.delta_star_p_tripped,
                                                  boundary_layer_data.delta_star_s_tripped,
                                                  alpha_star) 
        
        SPL_TBL_TE_untripped = TBL_TE_broadband_noise(f,r_er,L,U,M,R_c,Dbar_h,Dbar_l,Re_delta_star_p_untripped,
                                                  boundary_layer_data.delta_star_p_untripped,
                                                  boundary_layer_data.delta_star_s_untripped,
                                                  alpha_star)  
      
        # calculation of laminar boundary layer - vortex shedding, eqns 53 - 60 
        SPL_LBL_VS = LBL_VS_broadband_noise(R_c,alpha_star,boundary_layer_data.delta_star_p_untripped,r_er,L,M,Dbar_h,f,U)
       
        # calculation of tip vortex noise, eqns 61 - 67 
        alpha_TIP = abs(alpha_tip)
        SPL_TIP   = TIP_broadband_noise(alpha_TIP,M,c,c_0,f,Dbar_h,r_er)  
         
        # TO DO : Compute BWI  
        
        
        # TO DO : Compute BVI -  isnt this actually tonal ?
    
        # Unflatten Matices 
        SPL_TBL_TE_tripped   = unflatten_matrix(SPL_TBL_TE_tripped,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az) 
        SPL_TBL_TE_untripped = unflatten_matrix(SPL_TBL_TE_untripped,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        SPL_LBL_VS           = unflatten_matrix(SPL_LBL_VS,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        SPL_TIP              = unflatten_matrix(SPL_TIP,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        
        # Pressure from each Broadband source
        P_BWI                 = np.zeros((num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)) # this will be replaced soon
        P_TBL_TE_tripped      = 10**(SPL_TBL_TE_tripped/10)
        P_TBL_TE_untripped    = 10**(SPL_TBL_TE_untripped/10)
        P_LBL_VS              = 10**(SPL_LBL_VS/10)
        P_TIP                 = 10**(SPL_TIP/10)
        P_TIP[:,:,:,:-1,:,:]  = 0
        
        # Sum broadband compoments along blade sections and blades to get self noise per rotor 
        P_b_7                     = np.zeros((4,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az))
        P_b_7[0,:,:,:,:,:,:]      = P_BWI
        P_b_7[1,:,:,:,:,:,:]      = P_TBL_TE_tripped 
        P_b_7[2,:,:,:,:,:,:]      = P_LBL_VS
        P_b_7[3,:,:,:,:,:,:]      = P_TIP
        
        # Sum all components of broadband noise pressures
        P_b_6            = np.sum(P_b_7, axis=0)
        
        # Take product of total broadband noise with Doppler shift factor and weighting factor
        M_tot            = unflatten_matrix(M_tot,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        cos_zeta_r       = unflatten_matrix(cos_zeta_r,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az)
        Doppler_shift    = 1/(1-M_tot*cos_zeta_r)
        P_b_6_shifted    = (1/num_az)*(1/Doppler_shift)*P_b_6
        
        # Sum broadband noise along blade sections
        P_b_5            = np.sum(P_b_6_shifted, axis=3)
        
        # Sum broadband noise for all blades
        P_b_4            = np.sum(P_b_5, axis=2)
        
        # Sum broadband noise along all azimuthal stations
        P_b_3            = np.sum(P_b_4, axis=3)
        
        SPL_broadband    = 10*np.log10(P_b_3)
        
        # store results 
        Noise.SPL_prop_broadband_spectrum                   = SPL_broadband
        Noise.SPL_prop_broadband_1_3_spectrum               = Noise.SPL_prop_broadband_spectrum     # already in 1/3 octave spectrum
        
    return
 
def flatten_matrix(x,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az):
    return np.reshape(x,(num_cpt*num_mic*num_blades*num_sec*num_cf*num_az))
 
def unflatten_matrix(x,num_cpt,num_mic,num_blades,num_sec,num_cf,num_az):
    return np.reshape(x,(num_cpt,num_mic,num_blades,num_sec,num_cf,num_az))


# RCAIDE/Methods/Mission/Common/Update/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
  
from .linear_inertial_horizontal_position      import linear_inertial_horizontal_position
from .curvilinear_inertial_horizontal_position import curvilinear_inertial_horizontal_position
from .acceleration                             import acceleration
from .angular_acceleration                     import angular_acceleration
from .aerodynamics                             import aerodynamics
from .altitude                                 import altitude
from .atmosphere                               import atmosphere
from .energy                                   import energy
from .differentials_time                       import differentials_time
from .forces                                   import forces
from .noise                                    import noise
from .emissions                                import emissions   
from .moments                                  import moments
from .ground_forces                            import ground_forces
from .freestream                               import freestream
from .gravity                                  import gravity 
from .kinematics                               import kinematics
from .orientations                             import orientations
from .planet_position                          import planet_position
from .stability                                import stability
from .thrust                                   import thrust 
from .weights                                  import weights


# RCAIDE/Library/Plots/Energy/plot_solar_network_conditions.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style 
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_solar_network_conditions(results,
                    save_figure   = False,
                    show_legend   = True,
                    save_filename = "Solar_Flux",
                    file_type     = ".png",
                    width = 11, height = 7):
    """This plots the solar flux and power train performance of an solar powered aircraft

    Assumptions:
    None
     
    Source:
    None
    
    Outputs:
    Plots
    
    Properties Used:
    N/A
    """
    
    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))    

    fig = plt.figure(save_filename)
    fig.set_size_inches(width,height)
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))      
    axis_1 = plt.subplot(3,2,1)
    axis_2 = plt.subplot(3,2,2) 
    axis_3 = plt.subplot(3,2,3) 
    axis_4 = plt.subplot(3,2,4) 

    for network in results.segments[0].analyses.energy.vehicle.networks: 
        busses  = network.busses 
        for bus in busses:
            for b_i, battery in enumerate(bus.battery_modules):
                if b_i == 0 or bus.identical_battery_modules == False:
                    for i in range(len(results.segments)):  
                        bus_results         = results.segments[i].conditions.energy[bus.tag] 
                        time                = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min 
                        flux                = results.segments[i].conditions.energy.solar_flux[:,0]
                        charge              = bus_results.power_draw[:,0]
                        current             = bus_results.current_draw[:,0]
                        energy              = bus_results.energy[:,0] / Units.MJ 
                        axis_1 = plt.subplot(2,2,1)
                        if b_i == 0 and i ==0:              
                            axis_1.plot(time, flux, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width, label = battery.tag)
                        else:
                            axis_1.plot(time, flux, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
                        axis_1.set_ylabel(r'Solar Flux (W/m^2)')
                        set_axes(axis_1)    
                    
                        axis_2 = plt.subplot(2,2,2)
                        axis_2.plot(time, charge, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width) 
                        axis_2.set_ylabel(r'Charging Power (W)')
                        set_axes(axis_2) 
                    
                        axis_3 = plt.subplot(2,2,3)
                        axis_3.plot(time, current, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
                        axis_3.set_xlabel('Time (mins)')
                        axis_3.set_ylabel(r'Battery Current (A)')
                        set_axes(axis_3) 
                    
                        axis_4 = plt.subplot(2,2,4)
                        axis_4.plot(time, energy, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
                        axis_4.set_xlabel('Time (mins)')
                        axis_4.set_ylabel(r'Battery Energy (MJ)')
                        set_axes(axis_4)   
                            
    if show_legend:        
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4)  
    
    # Adjusting the sub-plots for legend 
    fig.tight_layout()
    fig.subplots_adjust(top=0.8)
    
    # set title of plot 
    title_text    = 'Solar Flux Conditions: ' + battery.tag  
    fig.suptitle(title_text)
    
    if save_figure:
        plt.savefig(save_filename + file_type)    
           
    return fig 


# RCAIDE/Library/Plots/Performance/Aerodynamics/plot_airfoil_polar_files.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 


# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------     
def plot_airfoil_polar_files(polar_data,
                             save_figure = False,
                             save_filename = "Airfoil_Polars",
                             file_type = ".png",
                             width = 11, height = 7):
    """This plots all airfoil polars in the list "airfoil_polar_paths" 

    Assumptions:
    None

    Source:
    None

    Inputs:
    airfoil_polar_paths   [list of strings]

    Outputs: 
    Plots

    Properties Used:
    N/A	
    """ 
  
    
    # get plotting style 
    ps      = plot_style()  
    
    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
    
    
    # Get raw data polars
    CL           = polar_data.lift_coefficients
    CD           = polar_data.drag_coefficients
    alpha        = polar_data.angle_of_attacks/Units.degrees
    Re_raw       = polar_data.reynolds_numbers
    n_Re         = len(polar_data.re_from_polar) 
        
     
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,n_Re))     
     
    fig   = plt.figure(save_filename)
    fig.set_size_inches(width,height) 
      
    for j in range(n_Re):
        
        Re_val = str(round(Re_raw[j])/1e6)+'e6'  
        
        axis_1 = plt.subplot(2,2,1)
        axis_1.plot(alpha, CL[j,:], color = line_colors[j], marker = ps.markers[0], linewidth = ps.line_width, label ='Re='+Re_val)
        axis_1.set_ylabel(r'$C_l$')
        axis_1.set_xlabel(r'$\alpha$')
        set_axes(axis_1)    
        
        axis_2 = plt.subplot(2,2,2)
        axis_2.plot(alpha,CD[j,:], color = line_colors[j], marker = ps.markers[0], linewidth = ps.line_width, label ='Re='+Re_val) 
        axis_2.set_ylabel(r'$C_d$')
        axis_2.set_xlabel(r'$\alpha$')
        set_axes(axis_2)  
        
        axis_3 = plt.subplot(2,2,3)
        axis_3.plot(CL[j,:],CD[j,:], color = line_colors[j], marker = ps.markers[0], linewidth = ps.line_width, label ='Re='+Re_val)
        axis_3.set_xlabel('$C_l$')
        axis_3.set_ylabel(r'$C_d$')
        set_axes(axis_3) 
    
        axis_4 = plt.subplot(2,2,4)
        axis_4.plot(alpha, CL[j,:]/CD[j,:], color = line_colors[j], marker = ps.markers[0], linewidth = ps.line_width, label ='Re='+Re_val) 
        axis_4.set_ylabel(r'$Cl/Cd$')
        axis_4.set_xlabel(r'$\alpha$')
        set_axes(axis_4)   
     
    # set title of plot 
    title_text    = 'Airfoil Polars'      
    fig.suptitle(title_text)
    
    if save_figure:
        plt.savefig(save_filename + file_type)   
    return fig


# RCAIDE/Methods/Noise/Common/atmospheric_attenuation.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# Python package imports   
import numpy as np  
    
# ----------------------------------------------------------------------------------------------------------------------  
#  Atmospheric Attenuation
# ----------------------------------------------------------------------------------------------------------------------      
def atmospheric_attenuation(dist,center_frequencies):
    """ This calculates a the atmospheric attenuation of sound for each frequency band
    as a function of the distance from the source to the observer using Average Atmospheric 
    Attenuation Rates from SAE-AIR-1845
    
    Assumptions:
       N/A

    Source:
       AEDT Technical Manual Table 11-1 (SAE-AIR-1845) 
        
    Inputs:
        dist - Array with the distance vector from the aircraft (source) to the 
            microphone position (observer)                                              [m]
        center_frequency - center frequencies of one-third octave band                  [Hz]

    Outputs: 
        delta_spl - The delta sound pressure level to be reduced from the lossless SPL 
        condition                                                                       [dB]
    
    Properties Used:
        None 
    """ 
    ctrl_pts  = len(dist)
    # Atmospheric attenuation factor for a 70% humidity and 25 Celsius at 1000ft - Based SAE model 
    Att_dB = np.array([0.033,0.033,0.033,0.066,0.066,0.098,0.131,0.131,
                       0.197,0.230,0.295,0.361,0.459,0.590,0.754,0.983,1.311,1.705,2.295,3.115,
                       3.607,5.246,7.213,9.836])
    if len(center_frequencies)>24: 
        no_Att_dB = np.zeros(len(center_frequencies)-24)
        Att_dB    = np.hstack((no_Att_dB,Att_dB)) 
    
    # Calculates de delta SPL as a function of the distance
    delta_spl = np.tile(Att_dB[None,:],(ctrl_pts,1))*(np.tile(dist[:,None],(1,len(Att_dB))))/100
    
    return delta_spl




# RCAIDE/Library/Compoments/Fuselage/Circle_Segment.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports    
from .Segment import  Segment

# ---------------------------------------------------------------------------------------------------------------------- 
#  Segment
# ----------------------------------------------------------------------------------------------------------------------   
class Circle_Segment(Segment):
    """
    A component representing a circular cross-sectional segment of a fuselage. Segments are used 
    to define the shape and dimensions of the fuselage through a series of connected cross-sections.

    Attributes
    ----------
    tag : str
        Identifier for the segment, defaults to 'segment'
        
    prev : Component
        Link to the previous segment in the fuselage chain, defaults to None
        
    next : Component
        Link to the next segment in the fuselage chain, defaults to None
        
    percent_x_location : float
        Longitudinal position as percentage of fuselage length, defaults to 0
        
    percent_y_location : float
        Lateral position as percentage of fuselage width, defaults to 0
        
    percent_z_location : float
        Vertical position as percentage of fuselage height, defaults to 0
        
    height : float
        Vertical dimension of the segment cross-section, defaults to 0
        
    width : float
        Lateral dimension of the segment cross-section, defaults to 0
        
    radius : float
        Shape parameter controlling cross-section corner rounding, defaults to 2

    Notes
    -----
    Segments are used to build up the complete fuselage geometry through a series of 
    cross-sections. Each segment's position is defined as a percentage of the overall 
    fuselage dimensions, allowing for flexible scaling and positioning.

    **Major Assumptions**
    
    * Segments are connected in sequence to form a continuous surface
    * Cross-sections are symmetric about the vertical plane
    * Transitions between segments are smooth
    
    **Definitions**

    'Cross-section'
        The 2D shape formed by intersecting the fuselage with a plane perpendicular 
        to its longitudinal axis
        
    'Curvature'
        Parameter controlling the smoothness of transition between vertical and 
        horizontal surfaces at the corners of the cross-section

    See Also
    --------
    RCAIDE.Library.Components.Fuselages.Fuselage
        Parent container for fuselage segments
    """

    def __defaults__(self): 
        """
        Sets default values for the fuselage segment attributes.
        """
        self.tag                     = 'super_ellipse_segment' 
          


# RCAIDE/Library/Plots/Performance/Aerodynamics/plot_rotor_disc_performance.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units

# python imports   
import matplotlib.pyplot as plt 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------      
def plot_rotor_disc_performance(prop,outputs,i=0,title=None,save_figure=False): 

    """Plots rotor disc performance

    Assumptions:
    None

    Source: 
    None
    
    Inputs
    outputs    - rotor outputs data structure 

    Outputs:
    Plots

    Properties Used:
    N/A
    
    """
     
    # Now plotting:
    psi  = outputs.disc_azimuthal_distribution[i,:,:]
    r    = outputs.disc_radial_distribution[i,:,:]
    psi  = np.append(psi,np.atleast_2d(np.ones_like(psi[:,0])).T*2*np.pi,axis=1)
    r    = np.append(r,np.atleast_2d(r[:,0]).T,axis=1)
    
    T    = outputs.disc_thrust_distribution[i]
    Q    = outputs.disc_torque_distribution[i]
    alf  = (outputs.disc_effective_angle_of_attack[i])/Units.deg
    va   = outputs.disc_axial_induced_velocity[i]
    vt   = outputs.disc_tangential_induced_velocity[i]
        
    
    T    = np.append(T,np.atleast_2d(T[:,0]).T,axis=1)
    Q    = np.append(Q,np.atleast_2d(Q[:,0]).T,axis=1)
    alf  = np.append(alf,np.atleast_2d(alf[:,0]).T,axis=1)
    
    va   = np.append(va, np.atleast_2d(va[:,0]).T, axis=1)
    vt   = np.append(vt, np.atleast_2d(vt[:,0]).T, axis=1)
    
    lev = 101
    cm  = 'jet'
    
    # plot the grid point velocities
    fig  = plt.figure(figsize=(12,8))
    ax0  = fig.add_subplot(231, polar=True)
    p0   = ax0.contourf(psi, r, T,lev,cmap=cm)
    ax0.set_title('Thrust Distribution',pad=15)      
    ax0.set_rorigin(0)
    ax0.set_yticklabels([])
    plt.colorbar(p0, ax=ax0)
     
    ax1  = fig.add_subplot(232, polar=True)   
    p1   = ax1.contourf(psi, r, Q,lev,cmap=cm) 
    ax1.set_title('Torque Distribution',pad=15) 
    ax1.set_rorigin(0)
    ax1.set_yticklabels([])    
    plt.colorbar(p1, ax=ax1)
     
    ax2  = fig.add_subplot(233, polar=True)       
    p2   = ax2.contourf(psi, r, alf,lev,cmap=cm) 
    ax2.set_title('Local Blade Angle (deg)',pad=15) 
    ax2.set_rorigin(0)
    ax2.set_yticklabels([])
    plt.colorbar(p2, ax=ax2)
 
    ax3  = fig.add_subplot(234, polar=True)       
    p3   = ax3.contourf(psi, r, va,lev,cmap=cm) 
    ax3.set_title('Axial Velocity',pad=15) 
    ax3.set_rorigin(0)
    ax3.set_yticklabels([])
    plt.colorbar(p3, ax=ax3)    
     
    ax4  = fig.add_subplot(235, polar=True)       
    p4   = ax4.contourf(psi, r, vt,lev,cmap=cm) 
    ax4.set_title('Tangential Velocity',pad=15) 
    ax4.set_rorigin(0)
    ax4.set_yticklabels([])
    plt.colorbar(p4, ax=ax4)    
  
    return fig 


# RCAIDE/Library/Compoments/Systems/Avionics.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
# RCAIDE imports  
from RCAIDE.Library.Components import Component
from RCAIDE.Library.Methods.Propulsors.Common.append_avionics_conditions import append_avionics_conditions
 
# ----------------------------------------------------------------------------------------------------------------------
#  Avionics
# ----------------------------------------------------------------------------------------------------------------------            
class Avionics(Component):
    """
    A class representing aircraft avionics systems and their power requirements.

    Attributes
    ----------
    power_draw : float
        Power consumption of the avionics system, defaults to 0.0
        
    tag : str
        Unique identifier for the avionics system, defaults to 'Avionics'

    Notes
    -----
    The avionics class models the electrical power requirements of aircraft 
    electronics and instruments. This includes:
    
    * Flight management systems
    * Navigation equipment
    * Communication systems
    * Display systems
    
    **Major Assumptions**
    
    * Constant power draw during operation
    * Instantaneous power availability
    * No thermal management considerations
    
    **Definitions**

    'Power Draw'
        The electrical power required by the avionics system during operation
        
    'Bus'
        The electrical distribution system supplying power to the avionics

    See Also
    --------
    RCAIDE.Library.Components.Systems.System
        Base system class
    """        
    def __defaults__(self):
        """
        Sets default values for the avionics system attributes.
        """                 
        self.power_draw = 0.0
        self.tag        = 'Avionics'

    def append_operating_conditions(self, segment, bus): 
        """
        Adds operating conditions for the avionics system to a mission segment.

        Parameters
        ----------
        segment : Data
            Mission segment to which conditions are being added
        bus : Data
            Electrical bus supplying power to the avionics
        """
        append_avionics_conditions(self, segment, bus)
        return
            
    def power(self):
        """
        Calculates the power draw from the avionics system.

        Returns
        -------
        float
            Power draw in Watts

        Notes
        -----
        Sets both the input power and returns the power draw value for use
        in energy calculations.
        """                 
        self.inputs.power = self.power_draw
        return self.power_draw


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/General_Aviation/compute_vertical_tail_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core import  Units,  Data
import  numpy as  np

# ----------------------------------------------------------------------------------------------------------------------
# Vertical Tail Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_vertical_tail_weight(S_v, AR_v, sweep_v, q_c, taper_v, t_c_v, Nult,TOW,t_tail, rudder_fraction = 0.25):      
    """
        Calculate the weight of the vertical fin of an aircraft without the weight of the rudder and then calculate the weight of the rudder        
        
        Source:
            Raymer,  Aircraft Design: A Conceptual Approach (pg 460 in 4th edition)
        
        Inputs:
            S_v - area of the vertical tail (combined fin and rudder)          [meters**2]
            M_w -mass of wing                                                  [kilograms]
            AR_v -aspect ratio of vertial tail                                 [dimensionless]
            sweep_v - sweep angle of the vertical tail                         [radians]
            q_c - dynamic pressure at cruise                                   [Pascals]
            taper_v - taper ratio of vertical tail                             [dimensionless]
            t_c_v -thickness to chord ratio of wing                            [dimensionless]
            Nult - ultimate load of the aircraft                               [dimensionless]
            TOW - maximum takeoff weight of the aircraft                       [kilograms]
            S_gross_w - wing gross area                                        [meters**2]
            t_tail - flag to determine if aircraft has a t-tail                [dimensionless]
            rudder_fraction - fraction of the vertical tail that is the rudder [dimensionless]
        
        Outputs:
            output - a dictionary with outputs:
                W_tail_vertical - weight of the vertical fin portion of the vertical tail [kilograms]
            
        Assumptions:
            Vertical tail weight is the weight of the vertical fin without the rudder weight.
           
   """     
    # unpack inputs
    W_0   = TOW / Units.lb # Convert kg to lbs
    S_vt  = S_v/ Units.ft**2 # Convert from meters squared to ft squared  
    q     = q_c /(Units.force_pound / Units.ft**2.)

    # Determine weight of the vertical portion of the tail
    if t_tail == "yes": 
        T_tail_factor = 1.# Weight of vertical portion of the T-tail is 25% more than a conventional tail
    else: 
        T_tail_factor = 0.

    # Calculate weight of wing for traditional aircraft vertical tail without rudder
    tail_vert_English = .073*(1+.2*T_tail_factor)*((Nult*W_0)**(.376))*(q**.122)*(S_vt**.873)*((100.*t_c_v/np.cos(sweep_v))**(-.49))*((AR_v/(np.cos(sweep_v)**2.))**.357)*(taper_v**.039)

    # packup outputs
    W_tail_vertical = tail_vert_English * Units.lbs # Convert from lbs to kg

    return W_tail_vertical



# RCAIDE/Methods/Energy/Sources/Battery/Aluminum_Air/compute_al_air_cell_performance.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  compute_al_air_cell_performance
# ----------------------------------------------------------------------------------------------------------------------  
def find_aluminum_mass(battery, energy):
    aluminum_mass = energy*battery.aluminum_mass_factor
    return aluminum_mass 

def find_water_mass(battery, energy):
    water_mass = energy*battery.water_mass_gain_factor
    return water_mass



# RCAIDE/Framework/Analyses/Emissions/Emission_Index_CRN_Method.py
# 
# Created:  Jul 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from RCAIDE.Framework.Core import  Data 
from RCAIDE.Framework.Analyses    import Process 
from RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method import * 
from RCAIDE.Framework.Analyses.Emissions            import Emissions 
  
import numpy as  np

# ----------------------------------------------------------------------------------------------------------------------
#  Correlation_Buildup
# ----------------------------------------------------------------------------------------------------------------------
class Emission_Index_CRN_Method(Emissions): 
    """ 
    Emissions Index Chemical Reactor Network Method
    """    
    
    def __defaults__(self): 
        """
        This sets the default values and methods for the analysis.
    
        Assumptions:
        None
    
        Source:
        None 
        """

        # conditions table, used for surrogate model training
        self.training                   = Data() 
        self.training.pressure          = np.linspace(10,30, 5) *1E6
        self.training.temperature       = np.linspace(700, 900, 5) 
        self.training.air_mass_flowrate = np.linspace(10, 60, 5) 
        self.training.fuel_to_air_ratio = np.linspace(0.01, 0.05, 5)        
        
        # surrogoate models                 
        self.surrogates                 = Data() 

        # build the evaluation process
        compute                         = Process()  
        compute.emissions               = None  
        self.process                    = Process()
        self.process.compute            = compute        
        
        return
    
    def initialize(self):
        
        """
        This function defines the analysis of the combustor emisisons with the CRN method.
        If a surrogate is selected, a function to train the data is executed, followed by
        the surrogate definition. Consequently, the emissions of different species is 
        retrieved from the surrogate Emission Indexes.
        If the surrogate is not required, the Emission Index is computed directly from the
        CRN method.
    
        Assumptions:
        None
    
        Source:
        None 
        """        
        
        use_surrogate   = self.settings.use_surrogate  

        # If we are using the surrogate
        if use_surrogate == True: 
            # sample training data
            train_CRN_EI_surrogates(self)

            # build surrogate
            build_CRN_EI_surrogates(self)  

        # build the evaluation process
        compute   =  self.process.compute                  
        if use_surrogate == True: 
            compute.emissions  = evaluate_CRN_emission_indices_surrogate
        else:
            compute.emissions  = evaluate_CRN_emission_indices_no_surrogate
        return 


    def evaluate(self,segment):
        """The default evaluate function.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        results   <RCAIDE data class>

        Properties Used:
        self.settings
        self.vehicle
        """          
        settings = self.settings
        vehicle  = self.vehicle 
        results  = self.process.compute(segment,settings,vehicle)

        return results


# RCAIDE/Methods/Aerodynamics/Airfoil_Panel_Method/panel_geometry.py
# 
# 
# Created:  Dec 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# pacakge imports  
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------
# panel_geometry
# ----------------------------------------------------------------------------------------------------------------------
def panel_geometry(x,y,npanel,ncases,ncpts):
    """Computes airfoil surface panelization parameters for later use in 
    the computation of the matrix of influence coefficients.        

    Assumptions:
    None

    Source:
    None 
                                                                       
    Inputs:                                                         
    x       -  Vector of x coordinates of the surface nodes  [unitless]         
    y       -  Vector of y coordinates of the surface nodes  [unitless]      
    npanel  -  Number of panels on the airfoil               [unitless]         
                                                                     
    Outputs:                                             
    l       -  Panel lengths                              [unitless]
    st      -  np.sin(theta) for each panel               [radians]
    ct      -  np.cos(theta) for each panel               [radians]
    xbar    -  x-coordinate of the midpoint of each panel [unitless]              
    ybar    -  y-coordinate of the midpoint of each panel [unitless]               
    
    
    Properties Used:
    N/A
    """     
    # compute various geometrical quantities    
    l    = np.sqrt((x[1:] -x[:-1])**2 +(y[1:] -y[:-1])**2)
    st   = (y[1:] -y[:-1])/l 
    ct   = (x[1:] -x[:-1])/l 
    xbar = (x[1:] +x[:-1])/2
    ybar = (y[1:] +y[:-1])/2 
    
    norm  = np.zeros((npanel,2,ncases,ncpts))
    norm[:,0,:,:]  =  -st
    norm[:,1,:,:]  =  ct 
    
    return l,st,ct,xbar,ybar,norm 



# RCAIDE/Library/Components/Energy/Sources/Battery_Modules/Lithium_Ion_LFP.py
# 
# 
# Created: Nov 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# RCAIDE imports 
import RCAIDE
from RCAIDE.Framework.Core          import Units,Data
from .Generic_Battery_Module import  Generic_Battery_Module
from RCAIDE.Library.Methods.Energy.Sources.Batteries.Lithium_Ion_LFP  import * 

# package imports 
import numpy as np  
import os
from scipy.interpolate  import NearestNDInterpolator

# ----------------------------------------------------------------------------------------------------------------------
#  Lithium_Ion_LFP
# ----------------------------------------------------------------------------------------------------------------------  
class Lithium_Ion_LFP(Generic_Battery_Module):
    """
    Class for modeling A123 26650 lithium iron phosphate battery characteristics
    
    Attributes
    ----------
    tag : str
        Identifier for the battery module (default: 'lithium_ion_lfp')
        
    power_split_ratio : float, optional
        Power distribution ratio for multiple battery systems
        
    number_of_cells : int
        Number of cells in the module (default: 1)
        
    maximum_energy : float
        Maximum energy storage capacity [J] (default: 0.0)
        
    maximum_power : float
        Maximum power output [W] (default: 0.0)
        
    maximum_voltage : float
        Maximum voltage output [V] (default: 0.0)
        
    cell : Data
        Cell-specific properties
        
        - chemistry : str
            Battery chemistry type (default: 'LiFePO4')
        - diameter : float
            Cell diameter [m] (default: 0.0185)
        - height : float
            Cell height [m] (default: 0.0653)
        - mass : float
            Cell mass [kg] (default: 0.03)
        - surface_area : float
            Total cell surface area [m^2]
        - volume : float
            Cell volume [m^3]
        - density : float
            Cell density [kg/m^3]
        - electrode_area : float
            Active electrode area [m^2] (default: 0.0342)
        - maximum_voltage : float
            Maximum cell voltage [V] (default: 3.6)
        - nominal_capacity : float
            Rated capacity [Ah] (default: 2.6)
        - nominal_voltage : float
            Nominal operating voltage [V] (default: 3.6)
        - resistance : float
            Internal resistance [Ohms] (default: 0.022)
        - specific_heat_capacity : float
            Cell specific heat [J/kgK] (default: 1115)
        - radial_thermal_conductivity : float
            Radial thermal conductivity [W/mK] (default: 0.475)
        - axial_thermal_conductivity : float
            Axial thermal conductivity [W/mK] (default: 37.6)
        - discharge_performance_map : NearestNDInterpolator
            Interpolator for voltage vs discharge characteristics

    Notes
    -----
    The LFP cell model includes detailed thermal and electrical characteristics
    based on the A123 26650 cell. Performance data is interpolated from
    experimental measurements.

    References
    ----------
    [1] LithiumWerks (2019).A123 26650 Datasheet
        https://a123batteries.com/product_images/uploaded_images/26650.pdf
        
    [2] Arora, S., & Kapoor, A. (2019). Experimental Study of Heat Generation 
        Rate during Discharge of LiFePO4 Pouch Cells. Batteries, 5(4), 70.
        https://doi.org/10.3390/batteries5040070

    See Also
    --------
    RCAIDE.Library.Components.Energy.Sources.Battery_Modules.Generic_Battery_Module
        Base battery module class
    """
    
    def __defaults__(self):
    
        # ----------------------------------------------------------------------------------------------------------------------
        #  Module Level Properties
        # ----------------------------------------------------------------------------------------------------------------------        
        self.tag                                               = 'lithium_ion_lfp' 
        self.power_split_ratio                                 = None
        self.number_of_cells                                   = 1
        self.maximum_energy                                    = 0.0
        self.maximum_power                                     = 0.0
        self.maximum_voltage                                   = 0.0       
        
        # ----------------------------------------------------------------------------------------------------------------------
        #  Cell Level Properties
        # ----------------------------------------------------------------------------------------------------------------------        
        self.cell.chemistry                   = 'LiFePO4'
        self.cell.diameter                    = 0.0185                                                    # [m]
        self.cell.height                      = 0.0653                                                    # [m]
        self.cell.mass                        = 0.03  * Units.kg                                          # [kg]
        self.cell.surface_area                = (np.pi*self.cell.height*self.cell.diameter) \
                                                + (0.5*np.pi*self.cell.diameter**2)                       # [m^2]

        self.cell.volume                      = np.pi*(0.5*self.cell.diameter)**2*self.cell.height        # [m^3] 
        self.cell.density                     = self.cell.mass/self.cell.volume                           # [kg/m^3]
        self.cell.electrode_area              = 0.0342                                                    # [m^2]  # estimated 
                                                        
        self.cell.maximum_voltage             = 3.6                                                       # [V]
        self.cell.nominal_capacity            = 2.6                                                       # [Amp-Hrs]
        self.cell.nominal_voltage             = 3.6                                                       # [V]
         
        self.cell.watt_hour_rating            = self.cell.nominal_capacity  * self.cell.nominal_voltage   # [Watt-hours]      
        self.cell.specific_energy             = self.cell.watt_hour_rating*Units.Wh/self.cell.mass        # [J/kg]
        self.cell.specific_power              = self.cell.specific_energy/self.cell.nominal_capacity      # [W/kg]   
        self.cell.resistance                  = 0.022                                                     # [Ohms]
                                                                                                            
        self.cell.specific_heat_capacity      = 1115                                                      # [J/kgK]                                                     
        self.cell.radial_thermal_conductivity = 0.475                                                     # [J/kgK]  
        self.cell.axial_thermal_conductivity  = 37.6                                                      # [J/kgK]  

        battery_raw_data                      = load_battery_results()                                                   
        self.cell.discharge_performance_map   = create_discharge_performance_map(battery_raw_data)

        return                                     

    def energy_calc(self,state,bus,coolant_lines, t_idx, delta_t): 
        """
        Computes the state of the LFP battery cell
        
        Parameters
        ----------
        state : Data
            Current system state
        bus : Component
            Connected electrical bus
        coolant_lines : Component
            Connected cooling system
        t_idx : int
            Time index
        delta_t : float
            Time step [s]
            
        Returns
        -------
        stored_results_flag : bool
            Flag indicating if results were stored
        stored_battery_tag : str
            Identifier for stored results
        """      
        stored_results_flag, stored_battery_tag =  compute_lfp_cell_performance(self,state,bus,coolant_lines, t_idx,delta_t) 
                        
        return stored_results_flag, stored_battery_tag
    
    def reuse_stored_data(self,state,bus,coolant_lines, t_idx, delta_t,stored_results_flag, stored_battery_tag):
        """
        Reuses previously stored battery performance data
        
        Parameters
        ----------
        state : Data
            Current system state
        bus : Component
            Connected electrical bus
        coolant_lines : Component
            Connected cooling system
        t_idx : int
            Time index
        delta_t : float
            Time step [s]
        stored_results_flag : bool
            Flag indicating stored results exist
        stored_battery_tag : str
            Identifier for stored results
        """
        reuse_stored_lfp_cell_data(self,state,bus,coolant_lines, t_idx, delta_t,stored_results_flag, stored_battery_tag)
        return    
      
    def update_battery_age(self,segment, battery_conditions,increment_battery_age_by_one_day): 
        """
        Updates battery age and degradation parameters
        
        Parameters
        ----------
        segment : Segment
            Flight segment data
        battery_conditions : Data
            Battery operating conditions
        increment_battery_age_by_one_day : bool
            Flag to increment battery age
        """
        update_lfp_cell_age(self,segment, battery_conditions,increment_battery_age_by_one_day)
        return 
    
def create_discharge_performance_map(raw_data):
    """
    Creates an interpolator for battery discharge voltage characteristics
    
    This function processes raw battery test data to create an interpolation
    function that predicts battery voltage based on C-rate, temperature,
    and discharge capacity.

    Parameters
    ----------
    raw_data : dict
        Dictionary containing battery test data with structure:
        {c_rate: {temperature: {'discharge': [...], 'voltage': [...]}}}

    Returns
    -------
    battery_data : NearestNDInterpolator
        Interpolator function that takes [C-rate, temperature, discharge_capacity]
        and returns voltage

    Notes
    -----
    The function creates a 3D interpolation of voltage as a function of:
    - C-rate (discharge current relative to capacity)
    - Temperature
    - Discharge capacity (state of charge)
    
    Uses nearest neighbor interpolation for computational efficiency. Linear
    interpolation is possible but increases computation time by ~30x.

    References
    ----------
    [1] Lin, X., Perez, H., Siegel, J. B., & Stefanopoulou, A. G. (2024).
        "An Electro-Thermal Model for the A123 26650 LiFePO4 Battery."
        University of Michigan.
        https://hdl.handle.net/2027.42/97341
    """    
    # Prepare lists for the data needed for interpolation
    c_rates = []
    temperatures = []
    discharge_capacities = []
    voltages = []

    # Iterate through the data structure to populate the lists
    for c_rate_key, temp_data in raw_data.items():
        c_rate = float(c_rate_key)  # Convert C-rate to float

        for temp_key, data in temp_data.items():
            initial_temp = int(temp_key.split()[-1])  # Extract temperature as an integer

            # Extend lists with the discharge, voltage, and other data
            discharge_capacities.extend(data['discharge'])
            voltages.extend(data['voltage'])
            c_rates.extend([c_rate] * len(data['discharge']))
            temperatures.extend([initial_temp] * len(data['discharge']))

    # Convert lists to numpy arrays
    points = np.array([c_rates, temperatures, discharge_capacities]).T
    values = np.array(voltages)

    # Create the interpolant
    battery_data = NearestNDInterpolator(points, values) # Can be replaced by a Linear Interpolator for a better fit but computation time increases by 30 times. 

    return battery_data

def load_battery_results(): 
    '''Load experimental raw data of LFP cells 
        
    Assumptions:
        
        
    Source:
        
    
    Args: 
        None
        
    Returns:
        battery_data: raw data from battery   [unitless]
    '''    
    current_dir = os.path.dirname(os.path.abspath(__file__))
    full_path = os.path.join(current_dir, 'lfp_raw_data.res')

    # Load the raw_data using RCAIDE.load()
    raw_data = RCAIDE.load(full_path)

    return raw_data


# RCAIDE/Methods/Noise/Correlation_Buildup/Airframe/trailing_edge_flap_noise.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

import numpy as np
from RCAIDE.Framework.Core import Units

# ----------------------------------------------------------------------------------------------------------------------  
# Compute the trailing edge flap noise
# ----------------------------------------------------------------------------------------------------------------------  
def trailing_edge_flap_noise(Sf,cf,deltaf,slots,velocity,M,phi,theta,distance,frequency):
    """ This calculates the noise from the flap trailing edge as a 1/3 octave band sound pressure level.

    Assumptions:
        Correlation based.
        
    Inputs:
        Sf         - Flap area                                                      [sq.ft]
        cf         - Flap chord                                                     [ft]
        deltaf     - Flap deflection                                                [rad]
        slots      - Number of slots                                                (Flap type)
        velocity   - Aircraft speed                                                 [kts]
        M          - Mach number                                                    [Unitless]
        phi        - Azimuthal angle                                                [rad]
        theta      - Polar angle                                                    [rad]
        distance   - Distance from airplane to observer, evaluated at retarded time [ft]
        frequency  - Frequency array                                                [Hz]
        
    Outputs: One Third Octave Band SPL                                              [dB]
        SPL                
 
    Properties Used:
        N/A 
    
    """

    # Process
    G      = np.zeros(24)

    test   = frequency*cf/(velocity/Units.ft*(1-M*np.cos(theta)))

    if (slots==1 or slots==2):
        for i in range (0,24):
            if (test[i]<2):
                G[i] = 99+10*np.log10(test[i])
            elif (test[i]<20):
                G[i] = 103.82-6*np.log10(test[i])
            else:
                G[i] = 135.04-30*np.log10(test[i])

    elif slots==3:
        for i in range(0,24):
            if(test[i]<2):
                G[i] = 99+10*np.log10(test[i])
            elif (test[i]<75):
                G[i] = 102.61-2*np.log10(test[i])
            else:
                G[i] = 158.11-30*np.log10(test[i])

    G = np.transpose(G)
    
    if theta+deltaf>=np.pi:
        directivity = 0.0
    else:     
        directivity = 20.0*np.log10(np.sin(theta)* (np.cos(phi))**2 * np.sin(theta+deltaf))

    SPL = G+10*np.log10(Sf*(np.sin(deltaf))**2/(distance**2))+ \
        60*np.log10((velocity/Units.kts)/100.0)+directivity

    return SPL 
